"""This module contains utility functions related to Enums."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum
from typing import List, Type


def get_enum_names(enum_cls: Type[Enum]) -> List[str]:
    """Returns a list containing the names of all members of the Enum class passed as `enum_cls`.

    Args:
        enum_cls (Type[Enum]): The Enum class to get the names of its members.

    Returns:
        List[str]: The list of names of all members of the Enum class passed as `enum_cls`.
    """
    return [member.name for member in enum_cls]



"""Tests the progress callback on a dummy lightning module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import pytorch_lightning as pl

from otx.algorithms.anomaly.adapters.anomalib.callbacks import ProgressCallback
from otx.api.entities.model_template import TaskType
from tests.unit.algorithms.anomaly.helpers.dummy_dataset import DummyDataModule
from tests.unit.algorithms.anomaly.helpers.dummy_model import DummyModel


class ProgressStageCheckerCallback(pl.Callback):
    """This callback is injected into the model to check the stage of progress callback.

    Args:
        progress_callback (ProgressCallback): Reference to progress callback.
    """

    def __init__(self, progress_callback: ProgressCallback):
        self.progress_callback = progress_callback

    def on_validation_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        assert self.progress_callback._get_progress("train") != 0.0

    def on_validation_batch_end(self, *args, **kwds) -> None:
        """Check that training progress is not 0.0 after validation batch end."""
        assert self.progress_callback._get_progress("train") != 0.0


class TestProgressCallback:
    def test_progress_callback(self):
        """Tests if progress callback runs and that the progress is not reset after validation step."""
        datamodule = DummyDataModule(TaskType.ANOMALY_CLASSIFICATION)
        model = DummyModel()
        progress_callback = ProgressCallback()
        # inject callback after progress callback
        stage_checker = ProgressStageCheckerCallback(progress_callback)
        # turn off sanity check on validation step as it will fail due to missing training data length
        trainer = pl.Trainer(
            logger=False,
            enable_checkpointing=False,
            max_epochs=5,
            callbacks=[progress_callback, stage_checker],
            num_sanity_val_steps=0,
        )
        trainer.fit(model, datamodule)
        trainer.test(model, datamodule)
        trainer.predict(model, datamodule)



"""Train Task of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
from typing import Any, Iterable, List, Optional

import numpy as np
import torch
from mmcv.utils import ConfigDict

from otx.algorithms.common.utils.callback import TrainingProgressCallback
from otx.algorithms.common.utils.data import get_dataset
from otx.algorithms.detection.adapters.mmdet.utils.config_utils import (
    should_cluster_anchors,
)
from otx.algorithms.detection.utils.data import adaptive_tile_params
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    CurveMetric,
    LineChartInfo,
    LineMetricsGroup,
    MetricsGroup,
    ScoreMetric,
    VisualizationType,
)
from otx.api.entities.model import ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

from .inference import DetectionInferenceTask

logger = get_logger()


# pylint: disable=too-many-locals, too-many-instance-attributes, too-many-ancestors
class DetectionTrainTask(DetectionInferenceTask, ITrainingTask):
    """Train Task Implementation of OTX Detection."""

    @check_input_parameters_type()
    def save_model(self, output_model: ModelEntity):
        """Save best model weights in DetectionTrainTask."""
        logger.info("called save_model")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "confidence_threshold": self.confidence_threshold,
            "VERSION": 1,
        }
        if self._model_cfg is not None and should_cluster_anchors(self._model_cfg):
            modelinfo["anchors"] = {}
            self._update_anchors(modelinfo["anchors"], self._model_cfg.model.bbox_head.anchor_generator)

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    def cancel_training(self):
        """Cancel training function in DetectionTrainTask.

        Sends a cancel training signal to gracefully stop the optimizer. The signal consists of creating a
        '.stop_training' file in the current work_dir. The runner checks for this file periodically.
        The stopping mechanism allows stopping after each iteration, but validation will still be carried out. Stopping
        will therefore take some time.
        """
        logger.info("Cancel training requested.")
        self._should_stop = True
        if self.cancel_interface is not None:
            self.cancel_interface.cancel()
        else:
            logger.info("but training was not started yet. reserved it to cancel")
            self.reserved_cancel = True

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: Optional[TrainParameters] = None,
    ):
        """Train function in DetectionTrainTask."""
        logger.info("train()")
        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTE LoggerHook & Time Monitor
        if train_parameters:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        self._data_cfg = self._init_train_data_cfg(dataset)
        self._is_training = True

        if bool(self._hyperparams.tiling_parameters.enable_tiling) and bool(
            self._hyperparams.tiling_parameters.enable_adaptive_params
        ):
            adaptive_tile_params(self._hyperparams.tiling_parameters, dataset)

        results = self._run_task(
            "DetectionTrainer",
            mode="train",
            dataset=dataset,
            parameters=train_parameters,
        )

        # Check for stop signal when training has stopped. If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            # output_model.model_status = ModelStatus.FAILED
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # get prediction on validation set
        self._is_training = False
        val_dataset = dataset.get_subset(Subset.VALIDATION)
        val_preds, val_map = self._infer_detector(val_dataset, InferenceParameters(is_evaluation=True))

        preds_val_dataset = val_dataset.with_empty_annotations()
        self._add_predictions_to_dataset(val_preds, preds_val_dataset, 0.0)

        result_set = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=val_dataset,
            prediction_dataset=preds_val_dataset,
        )

        # adjust confidence threshold
        if self._hyperparams.postprocessing.result_based_confidence_threshold:
            best_confidence_threshold = None
            logger.info("Adjusting the confidence threshold")
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=True)
            if metric.best_confidence_threshold:
                best_confidence_threshold = metric.best_confidence_threshold.value
            if best_confidence_threshold is None:
                raise ValueError("Cannot compute metrics: Invalid confidence threshold!")
            logger.info(f"Setting confidence threshold to {best_confidence_threshold} based on results")
            self.confidence_threshold = best_confidence_threshold
        else:
            metric = MetricsHelper.compute_f_measure(result_set, vary_confidence_threshold=False)

        # compose performance statistics
        # TODO[EUGENE]: HOW TO ADD A MAE CURVE FOR TaskType.COUNTING?
        performance = metric.get_performance()
        performance.dashboard_metrics.extend(
            DetectionTrainTask._generate_training_metrics(self._learning_curves, val_map)
        )
        logger.info(f"Final model performance: {str(performance)}")
        # save resulting model
        self.save_model(output_model)
        output_model.performance = performance
        # output_model.model_status = ModelStatus.SUCCESS
        logger.info("train done.")

    def _init_train_data_cfg(self, dataset: DatasetEntity):
        logger.info("init data cfg.")
        data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val", "unlabeled"],
            [Subset.TRAINING, Subset.VALIDATION, Subset.UNLABELED],
        ):
            subset = get_dataset(dataset, subset)
            if subset:
                data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        # Temparory remedy for cfg.pretty_text error
        for label in self._labels:
            label.hotkey = "a"
        return data_cfg

    @staticmethod
    def _generate_training_metrics(learning_curves, scores) -> Iterable[MetricsGroup[Any, Any]]:
        """Get Training metrics (epochs & scores).

        Parses the mmdetection logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []

        # Learning curves.
        for key, curve in learning_curves.items():
            len_x, len_y = len(curve.x), len(curve.y)
            if len_x != len_y:
                logger.warning(f"Learning curve {key} has inconsistent number of coordinates ({len_x} vs {len_y}.")
                len_x = min(len_x, len_y)
                curve.x = curve.x[:len_x]
                curve.y = curve.y[:len_x]
            metric_curve = CurveMetric(
                xs=np.nan_to_num(curve.x).tolist(),
                ys=np.nan_to_num(curve.y).tolist(),
                name=key,
            )
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(LineMetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))

        # Final mAP value on the validation set.
        output.append(
            BarMetricsGroup(
                metrics=[ScoreMetric(value=scores, name="mAP")],
                visualization_info=BarChartInfo("Validation score", visualization_type=VisualizationType.RADIAL_BAR),
            )
        )

        return output



"""Unit Tests for the MPA Dataset Pipelines Transforms Augments."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any

import pytest
from PIL import Image

from otx.mpa.modules.datasets.pipelines.transforms.augments import (
    Augments,
    CythonAugments,
)


@pytest.fixture
def image() -> Image.Image:
    return Image.new("RGB", (100, 100), color="red")


class TestAugment:
    @pytest.mark.parametrize(
        "augmentation_str, args",
        [
            ("autocontrast", {}),
            ("equalize", {}),
            ("solarize", {"threshold": 128}),
            ("posterize", {"bits_to_keep": 4}),
            ("posterize", {"bits_to_keep": 8}),
            ("color", {"factor": 0.5}),
            ("contrast", {"factor": 0.5}),
            ("brightness", {"factor": 0.5}),
            ("sharpness", {"factor": 0.5}),
            ("rotate", {"degree": 45}),
            ("shear_x", {"factor": 0.5}),
            ("shear_y", {"factor": 0.5}),
            ("translate_x_rel", {"pct": 0.5}),
            ("translate_y_rel", {"pct": 0.5}),
        ],
    )
    def test_augmentation_function(self, image: Image.Image, augmentation_str: str, args: dict[str, Any]) -> None:
        """Test that the augmentation functions returns an Image object."""
        augmentation_func = getattr(Augments, augmentation_str)
        result = augmentation_func(image, **args)
        assert isinstance(result, Image.Image)

    def test_rotate_with_list_interpolation_instance(self, image: Image.Image) -> None:
        """Test whether list of interpolation instances are accepted."""
        result = Augments.rotate(image, 45, resample=[Image.BICUBIC, Image.BILINEAR])
        assert isinstance(result, Image.Image)


class TestCythonAugments:
    @pytest.mark.parametrize(
        "augmentation_str, args",
        [
            ("autocontrast", {}),
            ("equalize", {}),
            ("solarize", {"threshold": 128}),
            ("posterize", {"bits_to_keep": 4}),
            ("posterize", {"bits_to_keep": 8}),
            ("color", {"factor": 0.5}),
            ("contrast", {"factor": 0.5}),
            ("brightness", {"factor": 0.5}),
            ("sharpness", {"factor": 0.5}),
        ],
    )
    def test_augmentation_output_equals_to_input_image(
        self, image: Image.Image, augmentation_str: str, args: dict[str, Any]
    ) -> None:
        """Test that the augmentation functions returns an Image object."""
        augmentation_func = getattr(CythonAugments, augmentation_str)
        result = augmentation_func(image, **args)
        assert isinstance(result, Image.Image)
        assert result == image

    @pytest.mark.parametrize(
        "augmentation_str, args",
        [
            ("rotate", {"degree": 45}),
            ("shear_x", {"factor": 0.5}),
            ("shear_y", {"factor": 0.5}),
            ("translate_x_rel", {"pct": 0.5}),
            ("translate_y_rel", {"pct": 0.5}),
        ],
    )
    def test_augmentation_output_not_equals_to_input_image(
        self, image: Image.Image, augmentation_str: str, args: dict[str, Any]
    ) -> None:
        """Test that the augmentation functions returns an Image object."""
        augmentation_func = getattr(CythonAugments, augmentation_str)
        result = augmentation_func(image, **args)
        assert isinstance(result, Image.Image)
        assert result != image

    def test_blend(self, image: Image.Image) -> None:
        """Test that it raises an assertion error if dst is not a numpy array."""
        with pytest.raises(AssertionError):
            CythonAugments.blend(image, image, 0.5)



"""Train Task of OTX Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
from typing import List, Optional

import torch
from mmcv.utils import ConfigDict

from otx.algorithms.common.utils.callback import TrainingProgressCallback
from otx.algorithms.common.utils.data import get_dataset
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.metrics import (
    CurveMetric,
    InfoMetric,
    LineChartInfo,
    MetricsGroup,
    Performance,
    ScoreMetric,
    VisualizationInfo,
    VisualizationType,
)
from otx.api.entities.model import ModelEntity
from otx.api.entities.subset import Subset
from otx.api.entities.train_parameters import TrainParameters, default_progress_callback
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

from .inference import SegmentationInferenceTask

logger = get_logger()


# pylint: disable=too-many-locals, too-many-instance-attributes, too-many-ancestors
class SegmentationTrainTask(SegmentationInferenceTask, ITrainingTask):
    """Train Task Implementation of OTX Segmentation."""

    @check_input_parameters_type()
    def save_model(self, output_model: ModelEntity):
        """Save best model weights in SegmentationTrainTask."""
        logger.info(f"called save_model: {self._model_ckpt}")
        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}
        model_ckpt = torch.load(self._model_ckpt)
        modelinfo = {
            "model": model_ckpt,
            "config": hyperparams_str,
            "labels": labels,
            "VERSION": 1,
        }

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        output_model.precision = self._precision

    def cancel_training(self):
        """Cancel training function in SegmentationTrainTask.

        Sends a cancel training signal to gracefully stop the optimizer. The signal consists of creating a
        '.stop_training' file in the current work_dir. The runner checks for this file periodically.
        The stopping mechanism allows stopping after each iteration, but validation will still be carried out. Stopping
        will therefore take some time.
        """
        logger.info("Cancel training requested.")
        self._should_stop = True
        if self.cancel_interface is not None:
            self.cancel_interface.cancel()
        else:
            logger.info("but training was not started yet. reserved it to cancel")
            self.reserved_cancel = True

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: Optional[TrainParameters] = None,
    ):
        """Train function in SegmentationTrainTask."""
        logger.info("train()")
        # Check for stop signal between pre-eval and training.
        # If training is cancelled at this point,
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # Set OTX LoggerHook & Time Monitor
        if train_parameters is not None:
            update_progress_callback = train_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback
        self._time_monitor = TrainingProgressCallback(update_progress_callback)

        self._data_cfg = self._init_train_data_cfg(dataset)
        self._is_training = True
        results = self._run_task("SegTrainer", mode="train", dataset=dataset, parameters=train_parameters)

        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled and no new
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        # get output model
        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            # output_model.model_status = ModelStatus.FAILED
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        # Get training metrics group from learning curves
        training_metrics, best_score = self._generate_training_metrics_group(self._learning_curves)
        performance = Performance(
            score=ScoreMetric(value=best_score, name=self.metric),
            dashboard_metrics=training_metrics,
        )

        logger.info(f"Final model performance: {str(performance)}")
        # save resulting model
        self.save_model(output_model)
        output_model.performance = performance
        # output_model.model_status = ModelStatus.SUCCESS
        self._is_training = False
        logger.info("train done.")

    def _init_train_data_cfg(self, dataset: DatasetEntity):
        logger.info("init data cfg.")
        data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val", "unlabeled"],
            [Subset.TRAINING, Subset.VALIDATION, Subset.UNLABELED],
        ):
            subset = get_dataset(dataset, subset)
            if subset:
                data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        # Temparory remedy for cfg.pretty_text error
        for label in self._labels:
            label.hotkey = "a"
        return data_cfg

    def _generate_training_metrics_group(self, learning_curves):
        """Get Training metrics (epochs & scores).

        Parses the mmsegmentation logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []
        # Model architecture
        architecture = InfoMetric(name="Model architecture", value=self._model_name)
        visualization_info_architecture = VisualizationInfo(
            name="Model architecture", visualisation_type=VisualizationType.TEXT
        )
        output.append(
            MetricsGroup(
                metrics=[architecture],
                visualization_info=visualization_info_architecture,
            )
        )
        # Learning curves
        best_score = -1
        for key, curve in learning_curves.items():
            metric_curve = CurveMetric(xs=curve.x, ys=curve.y, name=key)
            if key == f"val/{self.metric}":
                best_score = max(curve.y)
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(MetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))
        return output, best_score



"""ParameterGroup is the main class responsible for grouping configurable parameters together."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum
from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from otx.api.configuration.elements import metadata_keys
from otx.api.configuration.enums import AutoHPOState
from otx.api.configuration.enums.config_element_type import (
    ConfigElementType,
    ElementCategory,
)


@attr.s(auto_attribs=True, order=False, eq=False)
class ParameterGroup:
    """A group of configuration elements.

    Parameters living within the parameter group are typed attrs Attributes. The schema for each parameter is defined
    in its metadata, which can be retrieved using the `get_metadata` method from the parent ParameterGroup instance.

    Attributes:
        header (str): User friendly name for the parameter group, that will be
            displayed in the UI
        description (str): User friendly string describing what the parameter
            group represents, that will be displayed in the UI.
        visible_in_ui (bool): Boolean that controls whether or not this
            parameter group will be exposed through the REST API and
            shown in the UI. Set to False to hide this group. Defaults
            to True
    """

    header: str = attr.ib()
    description: str = attr.ib(default="Default parameter group description")
    visible_in_ui: bool = attr.ib(default=True)
    type: ConfigElementType = attr.ib(
        default=ConfigElementType.PARAMETER_GROUP,
        repr=False,
        init=False,
        on_setattr=attr.setters.frozen,
    )

    def __attrs_post_init__(self):
        """Update parameter and group after __init__.

        This method is called after the __init__ method to update the parameter and group fields of the ParameterGroup
        instance.
        """
        groups: List[str] = []
        parameters: List[str] = []
        self.__metadata_overrides: Dict[str, Any] = {}  # pylint:disable=attribute-defined-outside-init

        for attribute_or_method_name in dir(self):
            # Go over all attributes and methods of the class instance
            attribute_or_method = getattr(self, attribute_or_method_name)

            metadata = self.get_metadata(attribute_or_method_name)
            if metadata:
                # If the attribute or method has metadata, it might be a configurable parameter. In that case, check
                # its type to make sure it is one of the primitive parameter types
                _type = metadata.get("type", None)
                if _type is not None:
                    if _type.category == ElementCategory.PRIMITIVES:
                        # Add the parameter name to the `parameters` attribute
                        parameters.append(attribute_or_method_name)

            if isinstance(attribute_or_method, ParameterGroup):
                # Add the parameter group name to the `groups` attribute
                groups.append(attribute_or_method_name)
                # Also run the post_init for all groups in the group, in case of nested groups
                attribute_or_method.__attrs_post_init__()

        self.groups = groups  # pylint:disable=attribute-defined-outside-init
        self.parameters = parameters  # pylint:disable=attribute-defined-outside-init

    def get_metadata(self, parameter_name: str) -> dict:
        """Retrieve the metadata for a particular parameter from the group.

        Args:
            parameter_name (str): name of the parameter for which to get the
                metadata

        Returns:
            dict: dictionary containing the metadata for the requested
                parameter. Returns an empty dict if no metadata was found
                for the parameter, or if the parameter was not found in the
                group.
        """
        parameter = getattr(attr.fields(type(self)), parameter_name, None)
        if parameter is not None:
            parameter_metadata = getattr(parameter, "metadata", {})
            metadata_dict = dict(parameter_metadata)
            parameter_overrides = self.__metadata_overrides.get(parameter_name, None)
            if parameter_overrides is not None:
                for metadata_key, value_override in parameter_overrides.items():
                    metadata_dict.update({metadata_key: value_override})
            return metadata_dict
        return {}

    def set_metadata_value(
        self,
        parameter_name: str,
        metadata_key: str,
        value: Union[int, float, str, bool, Enum],
    ) -> bool:
        """Sets the value of a specific metadata item `metadata_key` for the parameter named `parameter_name`.

        Args:
            parameter_name (str): name of the parameter for which to get the
                metadata item
            metadata_key (str): name of the metadata value to set
            value (Union[int, float, str, bool, Enum]): New value to assign to the metadata item accessed by
                `metadata_key`. The type of `value` has to exactly match
                the type of the current value of the metadata item

        Returns:
            True if the metadata item was successfully updated, False
            otherwise
        """
        parameter = getattr(attr.fields(type(self)), parameter_name, None)
        if parameter is None:
            return False
        parameter_metadata = dict(getattr(parameter, "metadata", {}))
        if metadata_key not in metadata_keys.all_keys():
            return False
        metadata_value = parameter_metadata[metadata_key]
        if metadata_value is not None and type(metadata_value) is not type(
            value
        ):  # pylint: disable=unidiomatic-typecheck
            return False
        existing_overrides = self.__metadata_overrides.get(parameter_name, None)
        if existing_overrides is None:
            self.__metadata_overrides[parameter_name] = {metadata_key: value}
        else:
            existing_overrides.update({metadata_key: value})
        return True

    def update_auto_hpo_states(self):
        """Update hpo state based on teh value of parameters.

        Updates the `auto_hpo_state` metadata field for all parameters in the parameter
        group, based on the values of the parameters and the values of their
        `auto_hpo_value` metadata fields.
        """
        for parameter_name in self.parameters:
            metadata = self.get_metadata(parameter_name)
            if metadata[metadata_keys.AUTO_HPO_STATE] == AutoHPOState.NOT_POSSIBLE:
                continue
            auto_hpo_value = metadata[metadata_keys.AUTO_HPO_VALUE]
            if auto_hpo_value is None:
                continue
            if auto_hpo_value != getattr(self, parameter_name):
                auto_hpo_state = AutoHPOState.OVERRIDDEN
            else:
                auto_hpo_state = AutoHPOState.OPTIMIZED
            self.set_metadata_value(
                parameter_name=parameter_name,
                metadata_key=metadata_keys.AUTO_HPO_STATE,
                value=auto_hpo_state,
            )
        for group_name in self.groups:
            group = getattr(self, group_name)
            group.update_auto_hpo_states()

    def __eq__(self, other):
        """Comparison with support for dynamically generated ParameterGroups.

        Override default implementation of __eq__ to enable comparison of
        ParameterGroups generated dynamically via the config helper.
        """
        other_type = getattr(other, "type", None)
        if other_type == self.type:
            return self.__dict__ == other.__dict__
        return False


_ParameterGroup = TypeVar("_ParameterGroup", bound=ParameterGroup)


def add_parameter_group(group: Type[_ParameterGroup]) -> _ParameterGroup:
    """Wrapper to attr.ib to add nested parameter groups to a configuration."""
    return attr.ib(factory=group, type=group)



"""Anomaly Classification Task."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
from typing import Optional

import torch
from anomalib.models import AnomalyModule, get_model
from anomalib.post_processing import NormalizationMethod, ThresholdMethod
from anomalib.utils.callbacks import (
    MetricsConfigurationCallback,
    MinMaxNormalizationCallback,
    PostProcessingConfigurationCallback,
)
from pytorch_lightning import Trainer, seed_everything

from otx.algorithms.anomaly.adapters.anomalib.callbacks import ProgressCallback
from otx.algorithms.anomaly.adapters.anomalib.data import OTXAnomalyDataModule
from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask

from .inference import InferenceTask

logger = get_logger(__name__)


class TrainingTask(InferenceTask, ITrainingTask):
    """Base Anomaly Task."""

    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: TrainParameters,
        seed: Optional[int] = None,
    ) -> None:
        """Train the anomaly classification model.

        Args:
            dataset (DatasetEntity): Input dataset.
            output_model (ModelEntity): Output model to save the model weights.
            train_parameters (TrainParameters): Training parameters
            seed: (Optional[int]): Setting seed to a value other than 0 also marks PytorchLightning trainer's
                deterministic flag to True.
        """
        logger.info("Training the model.")

        config = self.get_config()

        if seed:
            logger.info(f"Setting seed to {seed}")
            seed_everything(seed, workers=True)
            config.trainer.deterministic = True

        logger.info("Training Configs '%s'", config)

        datamodule = OTXAnomalyDataModule(config=config, dataset=dataset, task_type=self.task_type)
        callbacks = [
            ProgressCallback(parameters=train_parameters),
            MinMaxNormalizationCallback(),
            MetricsConfigurationCallback(
                task=config.dataset.task,
                image_metrics=config.metrics.image,
                pixel_metrics=config.metrics.get("pixel"),
            ),
            PostProcessingConfigurationCallback(
                normalization_method=NormalizationMethod.MIN_MAX,
                threshold_method=ThresholdMethod.ADAPTIVE,
                manual_image_threshold=config.metrics.threshold.manual_image,
                manual_pixel_threshold=config.metrics.threshold.manual_pixel,
            ),
        ]

        self.trainer = Trainer(**config.trainer, logger=False, callbacks=callbacks)
        self.trainer.fit(model=self.model, datamodule=datamodule)

        self.save_model(output_model)

        logger.info("Training completed.")

    def load_model(self, otx_model: Optional[ModelEntity]) -> AnomalyModule:
        """Create and Load Anomalib Module from OTE Model.

        This method checks if the task environment has a saved OTE Model,
        and creates one. If the OTE model already exists, it returns the
        the model with the saved weights.

        Args:
            otx_model (Optional[ModelEntity]): OTE Model from the
                task environment.

        Returns:
            AnomalyModule: Anomalib
                classification or segmentation model with/without weights.
        """
        model = get_model(config=self.config)
        if otx_model is None:
            logger.info(
                "No trained model in project yet. Created new model with '%s'",
                self.model_name,
            )
        else:
            buffer = io.BytesIO(otx_model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))

            try:
                if model_data["config"]["model"]["backbone"] == self.config["model"]["backbone"]:
                    model.load_state_dict(model_data["model"])
                    logger.info("Loaded model weights from Task Environment")
                else:
                    logger.info(
                        "Model backbone does not match. Created new model with '%s'",
                        self.model_name,
                    )
            except BaseException as exception:
                raise ValueError("Could not load the saved model. The model file structure is invalid.") from exception

        return model



"""EfficientNet-V2 config for multi-class with contrastive loss for small datasets."""

# pylint: disable=invalid-name

_base_ = "../../base/models/efficientnet_v2.py"

model = dict(
    task="classification",
    type="SupConClassifier",
    backbone=dict(
        version="s_21k",
    ),
    head=dict(
        type="SupConClsHead",
        in_channels=-1,
        aux_mlp=dict(hid_channels=0, out_channels=1024),
        loss=dict(type="CrossEntropyLoss", loss_weight=1.0),
        aux_loss=dict(
            type="BarlowTwinsLoss",
            off_diag_penality=1.0 / 128.0,
            loss_weight=1.0,
        ),
    ),
)

fp16 = dict(loss_scale=512.0)



"""Base Self-SL dataset."""

__resize_target_size = 224
__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

__train_pipeline_v0 = [
    dict(type="RandomResizedCrop", size=__resize_target_size),
    dict(type="RandomFlip"),
    dict(
        type="RandomAppliedTrans",
        transforms=[
            dict(type="OTXColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", gray_prob=0.2),
    dict(type="GaussianBlur", sigma_min=0.1, sigma_max=2.0),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]
__train_pipeline_v1 = [
    dict(type="RandomResizedCrop", size=__resize_target_size),
    dict(type="RandomFlip"),
    dict(
        type="RandomAppliedTrans",
        transforms=[
            dict(type="OTXColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", gray_prob=0.2),
    dict(type="RandomAppliedTrans", transforms=[dict(type="GaussianBlur", sigma_min=0.1, sigma_max=2.0)], p=0.1),
    dict(type="RandomAppliedTrans", transforms=[dict(type="Solarize", thr=128)], p=0.2),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]

data = dict(
    samples_per_gpu=64,
    workers_per_gpu=2,
    train=dict(
        type="SelfSLDataset",
        pipeline=dict(
            view0=__train_pipeline_v0,
            view1=__train_pipeline_v1,
        ),
    ),
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List


def type_parser(graph, types) -> List[str]:
    found = []
    for node in graph:
        if node.type in types:
            found.append(node.name)
    return found


def result_parser(graph) -> List[str]:
    return type_parser(graph, ["Result"])


def parameter_parser(graph) -> List[str]:
    return type_parser(graph, ["Parameter"])



"""Configuration file of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=invalid-name

from attr import attrs

from otx.algorithms.common.configs import BaseConfig
from otx.api.configuration.elements import (
    add_parameter_group,
    boolean_attribute,
    configurable_boolean,
    configurable_integer,
    string_attribute,
)
from otx.api.configuration.enums import ModelLifecycle


@attrs
class ClassificationConfig(BaseConfig):
    """Configurations of classification task."""

    @attrs
    class __LearningParameters(BaseConfig.BaseLearningParameters):
        """Learning parameter configurations."""

        header = string_attribute("Learning Parameters")
        description = header

        max_num_epochs = configurable_integer(
            default_value=200,
            min_value=1,
            max_value=1000,
            header="Maximum number of training epochs",
            description="Increasing this value causes the results to be more robust but training time "
            "will be longer.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        enable_lr_finder = configurable_boolean(
            default_value=False,
            header="Enable automatic learing rate estimation",
            description="Learning rate parameter value will be ignored if enabled.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    @attrs
    class __AlgoBackend(BaseConfig.BaseAlgoBackendParameters):
        """Algorithm backend configurations."""

        header = string_attribute("Parameters for the MPA algo-backend")
        description = header

    @attrs
    class __POTParameter(BaseConfig.BasePOTParameter):
        """POT-related parameter configurations."""

        header = string_attribute("POT Parameters")
        description = header
        visible_in_ui = boolean_attribute(False)

    @attrs
    class __NNCFOptimization(BaseConfig.BaseNNCFOptimization):
        header = string_attribute("Optimization by NNCF")
        description = header
        visible_in_ui = boolean_attribute(False)

    learning_parameters = add_parameter_group(__LearningParameters)
    algo_backend = add_parameter_group(__AlgoBackend)
    pot_parameters = add_parameter_group(__POTParameter)
    nncf_optimization = add_parameter_group(__NNCFOptimization)



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import contextlib
import logging
import os
import random
import tempfile
from typing import Iterator, List, Optional, Sequence, Tuple

import cv2
import numpy as np
from bson import ObjectId

from otx.api.configuration import ConfigurableParameters
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle
from otx.api.entities.annotation import Annotation
from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle

logger = logging.getLogger(__name__)


def generate_unique_id() -> ID:
    """
    Generates unique ID for testing
    :return:
    """
    return ID(ObjectId())


class LabelSchemaExample:
    def __init__(self) -> None:
        self.label_domain = Domain.CLASSIFICATION

        self.flowering = self.new_label_by_name("flowering")
        self.no_plant = self.new_label_by_name("no_plant")
        self.vegetative = self.new_label_by_name("vegetative")

    def new_label_by_name(self, name: str, is_empty: bool = False) -> LabelEntity:
        label = LabelEntity(name=name, color=Color.random(), domain=self.label_domain, is_empty=is_empty)
        label.id_ = generate_unique_id()
        return label

    def add_hierarchy(self, label_schema: LabelSchemaEntity) -> Tuple[LabelEntity, LabelEntity, LabelEntity]:
        """Adds children to flowering, no_plant and vegetative"""
        label_schema.add_group(
            LabelGroup(
                "plant_state",
                [self.flowering, self.no_plant, self.vegetative],
                LabelGroupType.EXCLUSIVE,
            )
        )
        flower_partial_visible = self.new_label_by_name("flower_partial_visible")
        flower_fully_visible = self.new_label_by_name("flower_fully_visible")
        label_schema.add_group(
            LabelGroup(
                "flowering_state",
                [flower_fully_visible, flower_partial_visible],
                LabelGroupType.EXCLUSIVE,
            )
        )
        label_schema.add_child(self.flowering, flower_partial_visible)
        label_schema.add_child(self.flowering, flower_fully_visible)

        assert self.flowering == label_schema.get_parent(flower_partial_visible)
        assert label_schema.get_parent(self.no_plant) is None

        few_leaves = self.new_label_by_name("few_leaves")
        label_schema.add_group(LabelGroup("leaf_state", [few_leaves], LabelGroupType.EXCLUSIVE))
        label_schema.add_child(self.vegetative, few_leaves)
        return few_leaves, flower_fully_visible, flower_partial_visible


def generate_random_annotated_image(
    image_width: int,
    image_height: int,
    labels: Sequence[LabelEntity],
    min_size=50,
    max_size=250,
    shape: Optional[str] = None,
    max_shapes: int = 10,
    intensity_range: List[Tuple[int, int]] = None,
    random_seed: Optional[int] = None,
) -> Tuple[np.ndarray, List[Annotation]]:
    """
    Generate a random image with the corresponding annotation entities.

    :param intensity_range: Intensity range for RGB channels ((r_min, r_max), (g_min, g_max), (b_min, b_max))
    :param max_shapes: Maximum amount of shapes in the image
    :param shape: {"rectangle", "ellipse", "triangle"}
    :param image_height: Height of the image
    :param image_width: Width of the image
    :param labels: Task Labels that should be applied to the respective shape
    :param min_size: Minimum size of the shape(s)
    :param max_size: Maximum size of the shape(s)
    :param random_seed: Seed to initialize the random number generator
    :return: uint8 array, list of shapes
    """
    from skimage.draw import random_shapes, rectangle

    if intensity_range is None:
        intensity_range = [(100, 200)]

    image1: Optional[np.ndarray] = None
    sc_labels = []
    # Sporadically, it might happen there is no shape in the image, especially on low-res images.
    # It'll retry max 5 times until we see a shape, and otherwise raise a runtime error
    if shape == "ellipse":  # ellipse shape is not available in random_shapes function. use circle instead
        shape = "circle"
    for _ in range(5):
        rand_image, sc_labels = random_shapes(
            (image_height, image_width),
            min_shapes=1,
            max_shapes=max_shapes,
            intensity_range=intensity_range,
            min_size=min_size,
            max_size=max_size,
            shape=shape,
            random_seed=random_seed,
        )
        num_shapes = len(sc_labels)
        if num_shapes > 0:
            image1 = rand_image
            break

    if image1 is None:
        raise RuntimeError("Was not able to generate a random image that contains any shapes")

    annotations: List[Annotation] = []
    for sc_label in sc_labels:
        sc_label_name = sc_label[0]
        sc_label_shape_r = sc_label[1][0]
        sc_label_shape_c = sc_label[1][1]
        y_min, y_max = max(0.0, float(sc_label_shape_r[0] / image_height)), min(
            1.0, float(sc_label_shape_r[1] / image_height)
        )
        x_min, x_max = max(0.0, float(sc_label_shape_c[0] / image_width)), min(
            1.0, float(sc_label_shape_c[1] / image_width)
        )

        if sc_label_name == "ellipse":
            # Fix issue with newer scikit-image libraries that generate ellipses.
            # For now we render a rectangle on top of it
            sc_label_name = "rectangle"
            rr, cc = rectangle(
                start=(sc_label_shape_r[0], sc_label_shape_c[0]),
                end=(sc_label_shape_r[1] - 1, sc_label_shape_c[1] - 1),
                shape=image1.shape,
            )
            image1[rr, cc] = (
                random.randint(0, 200),  # nosec
                random.randint(0, 200),  # nosec
                random.randint(0, 200),  # nosec
            )
        if sc_label_name == "circle":
            sc_label_name = "ellipse"

        label_matches = [label for label in labels if sc_label_name == label.name]
        if len(label_matches) > 0:
            label = label_matches[0]
            box_annotation = Annotation(
                Rectangle(x1=x_min, y1=y_min, x2=x_max, y2=y_max),
                labels=[ScoredLabel(label, probability=1.0)],
            )

            annotation: Annotation

            if label.name == "ellipse":
                annotation = Annotation(
                    Ellipse(
                        x1=box_annotation.shape.x1,
                        y1=box_annotation.shape.y1,
                        x2=box_annotation.shape.x2,
                        y2=box_annotation.shape.y2,
                    ),
                    labels=box_annotation.get_labels(include_empty=True),
                )
            elif label.name == "triangle":
                points = [
                    Point(
                        x=(box_annotation.shape.x1 + box_annotation.shape.x2) / 2,
                        y=box_annotation.shape.y1,
                    ),
                    Point(x=box_annotation.shape.x1, y=box_annotation.shape.y2),
                    Point(x=box_annotation.shape.x2, y=box_annotation.shape.y2),
                ]

                annotation = Annotation(
                    Polygon(points=points),
                    labels=box_annotation.get_labels(include_empty=True),
                )
            else:
                annotation = box_annotation

            annotations.append(annotation)
        else:
            logger.warning(
                "Generated a random image, but was not able to associate a label with a shape. "
                f"The name of the shape was `{sc_label_name}`. "
            )

    return image1, annotations


@contextlib.contextmanager
def generate_random_image_folder(width: int = 480, height: int = 360, number_of_images: int = 10) -> Iterator[str]:
    """
    Generates a folder with random images, cleans up automatically if used in a `with` statement

    :param width: height of the images. Defaults to 480.
    :param height: width of the images. Defaults to 360.
    :param number_of_images: number of generated images. Defaults to 10.

    :return: The temporary directory
    """
    temp_dir = tempfile.TemporaryDirectory()

    for n in range(number_of_images):
        temp_file = os.path.join(temp_dir.name, f"{n}.jpg")
        _write_random_image(width, height, temp_file)

    try:
        yield temp_dir.name
    finally:
        temp_dir.cleanup()


@contextlib.contextmanager
def generate_random_video_folder(
    width: int = 480,
    height: int = 360,
    number_of_videos: int = 10,
    number_of_frames: int = 150,
) -> Iterator[str]:
    """
    Generates a folder with random videos, cleans up automatically if used in a `with` statement
    :param width: Width of the video. Defaults to 480.
    :param height: Height of the video. Defaults to 360.
    :param number_of_videos: Number of videos to generate. Defaults to 10.
    :param number_of_frames: Number of frames in each video. Defaults to 150.

    :return: A temporary directory with videos
    """
    temp_dir = tempfile.TemporaryDirectory()

    for n in range(number_of_videos):
        temp_file = os.path.join(temp_dir.name, f"{n}.mp4")
        _write_random_video(width, height, number_of_frames, temp_file)

    try:
        yield temp_dir.name
    finally:
        temp_dir.cleanup()


@contextlib.contextmanager
def generate_random_single_image(width: int = 480, height: int = 360) -> Iterator[str]:
    """
    Generates a random image, cleans up automatically if used in a `with` statement
    :param width: Width of the image. Defaults to 480.
    :param height: Height of the image. Defaults to 360.

    :return: Path to an image file
    """

    temp_dir = tempfile.TemporaryDirectory()
    temp_file = os.path.join(temp_dir.name, "temp_image.jpg")
    _write_random_image(width, height, temp_file)

    try:
        yield temp_file
    finally:
        temp_dir.cleanup()


@contextlib.contextmanager
def generate_random_single_video(width: int = 480, height: int = 360, number_of_frames: int = 150) -> Iterator[str]:
    """
    Generates a random video, cleans up automatically if used in a `with` statement
    :param width: Width of the video. Defaults to 480.
    :param height: Height of the video. Defaults to 360.
    :param number_of_frames: Number of frames in the video. Defaults to 150.

    :return: Path to a video file
    """
    temp_dir = tempfile.TemporaryDirectory()
    temp_file = os.path.join(temp_dir.name, "temp_video.mp4")
    _write_random_video(width, height, number_of_frames, temp_file)

    try:
        yield temp_file
    finally:
        temp_dir.cleanup()


def _write_random_image(width: int, height: int, filename: str):
    img = np.uint8(np.random.random((height, width, 3)) * 255)
    cv2.imwrite(filename, img)


def _write_random_video(width: int, height: int, number_of_frames: int, filename: str):
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    f = filename
    videowriter = cv2.VideoWriter(f, fourcc, 30, (width, height))

    for _ in range(number_of_frames):
        img = np.uint8(np.random.random((height, width, 3)) * 255)
        videowriter.write(img)

    videowriter.release()


class ConfigExample(ConfigurableParameters):
    header = string_attribute("Test configuration for an object detection task")
    description = header

    class __LearningParameters(ParameterGroup):
        header = string_attribute("Test Learning Parameters")
        description = header

        batch_size = configurable_integer(
            default_value=5,
            min_value=1,
            max_value=512,
            header="Test batch size",
            description="The number of training samples seen in each iteration of training. Increasing this value "
            "improves training time and may make the training more stable. A larger batch size has higher "
            "memory requirements.",
            warning="Increasing this value may cause the system to use more memory than available, "
            "potentially causing out of memory errors, please update with caution.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        num_iters = configurable_integer(
            default_value=1,
            min_value=1,
            max_value=100000,
            header="Number of training iterations",
            description="Increasing this value causes the results to be more robust but training time will be longer.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        learning_rate = configurable_float(
            default_value=0.01,
            min_value=1e-07,
            max_value=1e-01,
            header="Learning rate",
            description="Increasing this value will speed up training convergence but might make it unstable.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        learning_rate_warmup_iters = configurable_integer(
            default_value=100,
            min_value=1,
            max_value=10000,
            header="Number of iterations for learning rate warmup",
            description="Test learning rate warmup",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

        num_workers = configurable_integer(
            default_value=4,
            min_value=2,
            max_value=10,
            header="num_workers test header",
            description="num_workers test description",
            affects_outcome_of=ModelLifecycle.NONE,
        )

    class __Postprocessing(ParameterGroup):
        header = string_attribute("Test Postprocessing")
        description = header

        result_based_confidence_threshold = configurable_boolean(
            default_value=True,
            header="Test Result based confidence threshold",
            description="Test confidence threshold is derived from the results",
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )

        confidence_threshold = configurable_float(
            default_value=0.25,
            min_value=0,
            max_value=1,
            header="Test Confidence threshold",
            description="This threshold only takes effect if the threshold is not set based on the result.--Only test",
            affects_outcome_of=ModelLifecycle.INFERENCE,
        )

    learning_parameters = add_parameter_group(__LearningParameters)
    postprocessing = add_parameter_group(__Postprocessing)



"""Instance segmentation models static deploy config."""

_base_ = ["./base_detection_static.py"]

ir_config = dict(
    output_names=[
        "boxes",
        "labels",
        "masks",
    ]
)

codebase_config = dict(
    post_processing=dict(
        export_postprocess_mask=False,
    )
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Sequence, Union

from mmcv.runner import HOOKS, Hook
from torch.utils.data import DataLoader

from otx.mpa.modules.datasets.composed_dataloader import ComposedDL
from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class ComposedDataLoadersHook(Hook):
    def __init__(
        self,
        data_loaders: Union[Sequence[DataLoader], DataLoader],
    ):
        self.data_loaders = []
        self.composed_loader = None

        self.add_dataloaders(data_loaders)

    def add_dataloaders(self, data_loaders: Union[Sequence[DataLoader], DataLoader]):
        if isinstance(data_loaders, DataLoader):
            data_loaders = [data_loaders]
        else:
            data_loaders = list(data_loaders)

        self.data_loaders.extend(data_loaders)
        self.composed_loader = None

    def before_epoch(self, runner):
        if self.composed_loader is None:
            logger.info(
                "Creating ComposedDL "
                f"(runner's -> {runner.data_loader}, "
                f"hook's -> {[i for i in self.data_loaders]})"
            )
            self.composed_loader = ComposedDL([runner.data_loader, *self.data_loaders])
        # Per-epoch replacement: train-only loader -> train loader + additional loaders
        # (It's similar to local variable in epoch. Need to update every epoch...)
        runner.data_loader = self.composed_loader



"""Sample code of otx training for classification."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=too-many-locals, invalid-name, too-many-statements

import argparse
import random
import sys

import numpy as np
import torch
from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import (
    LabelEntity,
    LabelGroup,
    LabelGroupType,
    LabelSchemaEntity,
)
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

SEED = 5
random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed_all(SEED)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

logger = get_logger(name="mmcls")

parser = argparse.ArgumentParser(description="Sample showcasing the new API")
parser.add_argument("template_file_path", help="path to template file")
parser.add_argument("--export", action="store_true")
parser.add_argument("--multilabel", action="store_true")
parser.add_argument("--hierarchical", action="store_true")

args = parser.parse_args()


def load_test_dataset(data_type):
    """Load test dataset."""
    import PIL
    from PIL import ImageDraw

    from otx.api.entities.annotation import (
        Annotation,
        AnnotationSceneEntity,
        AnnotationSceneKind,
    )
    from otx.api.entities.dataset_item import DatasetItemEntity
    from otx.api.entities.image import Image
    from otx.api.entities.scored_label import ScoredLabel
    from otx.api.entities.shapes.rectangle import Rectangle

    def gen_image(resolution, shape=None):
        image = PIL.Image.new("RGB", resolution, (255, 255, 255))
        draw = ImageDraw.Draw(image)
        h, w = image.size
        shape = shape.split("+") if "+" in shape else [shape]
        for s in shape:
            if s == "rectangle":
                draw.rectangle((h * 0.1, w * 0.1, h * 0.4, w * 0.4), fill=(0, 192, 192), outline=(0, 0, 0))
            if s == "triangle":
                draw.polygon(
                    ((h * 0.5, w * 0.25), (h, w * 0.25), (h * 0.8, w * 0.5)), fill=(255, 255, 0), outline=(0, 0, 0)
                )
            if s == "pieslice":
                draw.pieslice(
                    ((h * 0.1, w * 0.5), (h * 0.5, w * 0.9)), start=50, end=250, fill=(0, 255, 0), outline=(0, 0, 0)
                )
            if s == "circle":
                draw.ellipse((h * 0.5, w * 0.5, h * 0.9, w * 0.9), fill="blue", outline="blue")
            if s == "text":
                draw.text((0, 0), "Intel", fill="blue", align="center")
        return np.array(image), shape

    datas = [
        gen_image((32, 32), shape="rectangle"),
        gen_image((32, 32), shape="triangle"),
        gen_image((32, 32), shape="rectangle+triangle"),  # for multilabel (old)
        gen_image((32, 32), shape="pieslice"),
        gen_image((32, 32), shape="pieslice+rectangle"),
        gen_image((32, 32), shape="pieslice+triangle"),
        gen_image((32, 32), shape="pieslice+rectangle+triangle"),  # for multilabel (new)
        gen_image((32, 32), shape="circle"),
        gen_image((32, 32), shape="circle+text"),  # for hierarchical (new)
    ]

    labels = {
        "rectangle": LabelEntity(name="rectangle", domain=Domain.CLASSIFICATION, id=0),
        "triangle": LabelEntity(name="triangle", domain=Domain.CLASSIFICATION, id=1),
        "pieslice": LabelEntity(name="pieslice", domain=Domain.CLASSIFICATION, id=2),
        "circle": LabelEntity(name="circle", domain=Domain.CLASSIFICATION, id=3),
        "text": LabelEntity(name="text", domain=Domain.CLASSIFICATION, id=4),
    }

    def get_image(i, subset, ignored_labels=None):
        image, shape = datas[i]
        lbl = [ScoredLabel(label=labels[s], probability=1.0) for s in shape]
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[
                    Annotation(
                        Rectangle(x1=0.0, y1=0.0, x2=1.0, y2=1.0),
                        labels=lbl,
                    )
                ],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
            ignored_labels=ignored_labels,
        )

    def gen_old_new_dataset(multilabel=False, hierarchical=False):
        old_train, old_val, new_train, new_val = [], [], [], []
        old_repeat = 8
        new_repeat = 4
        if multilabel:
            old_img_idx = [0, 1, 2]
            new_img_idx = [0, 1, 2, 3, 4, 5, 6]
            ignored_labels = [labels["pieslice"]]
        elif hierarchical:
            old_img_idx = [0, 1, 2, 3, 8]
            new_img_idx = [0, 1, 2, 3, 8, -1]
            ignored_labels = [labels["text"]]
        else:
            old_img_idx = [0, 1]
            new_img_idx = [0, 1, 3]

        for _ in range(old_repeat):
            for idx in old_img_idx:
                old_train.append(get_image(idx, Subset.TRAINING))
                old_val.append(get_image(idx, Subset.VALIDATION))
        for _ in range(new_repeat):
            for idx in new_img_idx:
                if multilabel:
                    new_train.append(get_image(idx, Subset.TRAINING, ignored_labels=ignored_labels))
                elif hierarchical:
                    new_train.append(get_image(idx, Subset.TRAINING, ignored_labels=ignored_labels))
                else:
                    new_train.append(get_image(idx, Subset.TRAINING))
                new_val.append(get_image(idx, Subset.VALIDATION))

        return old_train + old_val, new_train + new_val

    old, new = gen_old_new_dataset(args.multilabel, args.hierarchical)

    if not args.hierarchical:
        labels = [labels["rectangle"], labels["triangle"], labels["pieslice"]]
    else:
        labels = list(labels.values())

    if data_type == "old":
        return DatasetEntity(old), labels[:-1]
    return DatasetEntity(old + new), labels


def get_label_schema(labels, multilabel=False, hierarchical=False):
    """Get label schema."""
    label_schema = LabelSchemaEntity()

    emptylabel = LabelEntity(id=-1, name="Empty label", is_empty=True, domain=Domain.CLASSIFICATION)
    empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)

    if multilabel:
        for label in labels:
            label_schema.add_group(LabelGroup(name=label.name, labels=[label], group_type=LabelGroupType.EXCLUSIVE))
        label_schema.add_group(empty_group)
    elif hierarchical:
        single_label_classes = ["pieslice", "circle"]
        multi_label_classes = ["rectangle", "triangle", "text"]

        single_labels = [label for label in labels if label.name in single_label_classes]
        single_label_group = LabelGroup(name="labels", labels=single_labels, group_type=LabelGroupType.EXCLUSIVE)
        label_schema.add_group(single_label_group)

        for label in labels:
            if label.name in multi_label_classes:
                label_schema.add_group(
                    LabelGroup(
                        name=f"{label.name}____{label.name}_group", labels=[label], group_type=LabelGroupType.EXCLUSIVE
                    )
                )
    else:
        main_group = LabelGroup(name="labels", labels=labels, group_type=LabelGroupType.EXCLUSIVE)
        label_schema.add_group(main_group)

    return label_schema


def validate(task, validation_dataset, model):
    """Validate."""
    print("Get predictions on the validation set")
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
    )
    resultset = ResultSetEntity(
        model=model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    print("Estimate quality on validation set")
    task.evaluate(resultset)
    print(str(resultset.performance))


def main():
    """Main of Classification Sample Test."""

    logger.info("Train initial model with OLD dataset")
    dataset, labels_list = load_test_dataset("old")
    labels_schema = get_label_schema(labels_list, multilabel=args.multilabel, hierarchical=args.hierarchical)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 10
    params.learning_parameters.learning_rate = 0.03
    params.learning_parameters.learning_rate_warmup_iters = 4
    params.learning_parameters.batch_size = 16

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    initial_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, initial_model)

    logger.info("Class-incremental learning with OLD + NEW dataset")
    dataset, labels_list = load_test_dataset("new")
    labels_schema = get_label_schema(labels_list, multilabel=args.multilabel, hierarchical=args.hierarchical)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 10
    params.learning_parameters.learning_rate = 0.03
    params.learning_parameters.learning_rate_warmup_iters = 4
    params.learning_parameters.batch_size = 16

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=initial_model, hyper_parameters=params, label_schema=labels_schema, model_template=model_template
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the validation set")
    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export or args.multilabel or args.hierarchical:
        logger.info("Export model")
        exported_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.export(ExportType.OPENVINO, exported_model)

        logger.info("Create OpenVINO Task")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        openvino_task_cls = get_task_class(openvino_task_impl_path)
        openvino_task = openvino_task_cls(environment)

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )

        resultset = ResultSetEntity(
            model=exported_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Estimate quality on validation set")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        # POT test
        logger.info("Run POT optimization")
        optimized_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        openvino_task.optimize(OptimizationType.POT, dataset, optimized_model, OptimizationParameters())
        logger.info("Run POT deploy")
        openvino_task.deploy(optimized_model)
        validate(task, validation_dataset, optimized_model)

        # NNCF test
        task_impl_path = model_template.entrypoints.nncf
        nncf_task_cls = get_task_class(task_impl_path)

        print("Create NNCF Task")
        environment.model = output_model
        task = nncf_task_cls(task_environment=environment)

        print("Optimize model by NNCF")
        optimized_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.optimize(OptimizationType.NNCF, dataset, optimized_model, None)
        validate(task, validation_dataset, output_model)


if __name__ == "__main__":
    sys.exit(main() or 0)



"""This module contains the implementations of performance providers for multi-score anomaly metrics."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import ABC
from typing import List, Optional

from otx.api.entities.metrics import (
    MetricsGroup,
    MultiScorePerformance,
    Performance,
    ScoreMetric,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.averaging import MetricAverageMethod
from otx.api.usecases.evaluation.dice import DiceAverage
from otx.api.usecases.evaluation.f_measure import FMeasure
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)
from otx.api.utils.dataset_utils import (
    contains_anomalous_images,
    split_local_global_resultset,
)


class AnomalyLocalizationPerformance(MultiScorePerformance):
    """Anomaly specific MultiScorePerformance.

    This class implements a special case of the MultiScorePerformance, specific for anomaly tasks that perform
    anomaly localization (detection/segmentation), in addition to anomaly classification.

    Args:
        global_score: Image-level performance metric.
        local_score: Pixel- or bbox-level performance metric, depending
            on the task type.
        dashboard_metrics: (optional) additional statistics, containing
            charts, curves, and other additional info.
    """

    def __init__(
        self,
        global_score: ScoreMetric,
        local_score: Optional[ScoreMetric],
        dashboard_metrics: Optional[List[MetricsGroup]],
    ):
        super().__init__(
            primary_score=local_score,
            additional_scores=[global_score],
            dashboard_metrics=dashboard_metrics,
        )
        self._global_score = global_score
        self._local_score = local_score

    @property
    def global_score(self):
        """Return the global (image-level) score metric."""
        return self._global_score

    @property
    def local_score(self):
        """Return the local (pixel-/bbox-level) score metric."""
        return self._local_score


class AnomalyLocalizationScores(IPerformanceProvider, ABC):
    """AnomalyLocalizationPerformance object for anomaly segmentation and anomaly detection tasks.

    Depending on the subclass, the `get_performance` method returns an AnomalyLocalizationPerformance object with the
    pixel- or bbox-level metric as the primary score. The global (image-level) performance metric is included as an
    additional metric.

    Args:
        resultset: ResultSet that scores will be computed for
    """

    def __init__(self, resultset: ResultSetEntity):
        self.local_score: Optional[ScoreMetric] = None
        self.dashboard_metrics: List[MetricsGroup] = []

        global_resultset, local_resultset = split_local_global_resultset(resultset)

        global_metric = FMeasure(resultset=global_resultset)
        global_performance = global_metric.get_performance()
        self.global_score = global_performance.score
        self.dashboard_metrics += global_performance.dashboard_metrics

        if contains_anomalous_images(local_resultset.ground_truth_dataset):
            local_metric = self._get_local_metric(local_resultset)
            local_performance = local_metric.get_performance()
            self.local_score = local_performance.score
            self.dashboard_metrics += local_performance.dashboard_metrics

    @staticmethod
    def _get_local_metric(local_resultset: ResultSetEntity) -> IPerformanceProvider:
        """Return the local performance metric for the resultset."""
        raise NotImplementedError

    def get_performance(self) -> Performance:
        """Return the performance object for the resultset."""
        return AnomalyLocalizationPerformance(
            global_score=self.global_score,
            local_score=self.local_score,
            dashboard_metrics=self.dashboard_metrics,
        )


class AnomalySegmentationScores(AnomalyLocalizationScores):
    """Performance provider for anomaly segmentation tasks."""

    @staticmethod
    def _get_local_metric(local_resultset: ResultSetEntity) -> IPerformanceProvider:
        return DiceAverage(resultset=local_resultset, average=MetricAverageMethod.MICRO)


class AnomalyDetectionScores(AnomalyLocalizationScores):
    """Performance provider for anomaly detection tasks."""

    @staticmethod
    def _get_local_metric(local_resultset: ResultSetEntity) -> IPerformanceProvider:
        return FMeasure(resultset=local_resultset)



_base_ = [
    "../_base_/logs/tensorboard_logger.py",
    "../_base_/optimizers/sgd.py",
    "../_base_/runners/epoch_runner.py",
    "../_base_/data/selfsl_seg_data.py",
]


# doesn't inherit default.py to disenable task_adapt
cudnn_benchmark = True

seed = 5
deterministic = False

hparams = dict(dummy=0)

optimizer = dict(type="SGD", lr=0.001, momentum=0.9, weight_decay=10e-4)

lr_config = dict(policy="customstep", by_epoch=True, gamma=1, step=10)

optimizer_config = dict(_delete_=True, grad_clip=dict(max_norm=40, norm_type=2))

log_config = dict(
    interval=1,
    hooks=[
        dict(type="TextLoggerHook", by_epoch=True, ignore_last=False),
    ],
)

dist_params = dict(backend="nccl", linear_scale_lr=False)

runner = dict(type="EpochBasedRunner", max_epochs=10)

checkpoint_config = dict(by_epoch=True, interval=1)

find_unused_parameters = False



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
from mmcv import Config

from otx.algorithms.common.adapters.mmcv.utils import (
    config_from_string,
    get_data_cfg,
    is_epoch_based_runner,
    patch_color_conversion,
    prepare_for_testing,
    prepare_work_dir,
    remove_from_config,
)
from otx.api.entities.datasets import DatasetEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)

# TODO: Need to add {patch_data_pipeline, get_meta_keys} unit-test


class TestMMCVUtilsInputParamsValidation:
    @e2e_pytest_unit
    def test_is_epoch_based_runner_input_params_validation(self):
        """
        <b>Description:</b>
        Check "is_epoch_based_runner" function input parameters validation

        <b>Input data:</b>
        "runner_config" non-ConfigDict object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "runner_config" function
        """
        with pytest.raises(ValueError):
            is_epoch_based_runner(runner_config="unexpected_str")  # type: ignore

    @e2e_pytest_unit
    def test_prepare_for_testing_input_params_validation(self):
        """
        <b>Description:</b>
        Check "prepare_for_testing" function input parameters validation

        <b>Input data:</b>
        "prepare_for_testing" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "prepare_for_testing" function
        """
        correct_values_dict = {"config": Config(), "dataset": DatasetEntity()}
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "config" parameter
            ("config", unexpected_str),
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=prepare_for_testing,
        )

    @e2e_pytest_unit
    def test_config_from_string_input_params_validation(self):
        """
        <b>Description:</b>
        Check "config_from_string" function input parameters validation

        <b>Input data:</b>
        "config_string" non-string type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "config_from_string" function
        """
        with pytest.raises(ValueError):
            config_from_string(config_string=1)  # type: ignore

    @e2e_pytest_unit
    def test_prepare_work_dir_input_params_validation(self):
        """
        <b>Description:</b>
        Check "prepare_work_dir" function input parameters validation

        <b>Input data:</b>
        "config" non-Config type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "prepare_work_dir" function
        """
        with pytest.raises(ValueError):
            prepare_work_dir(config=1)  # type: ignore

    @e2e_pytest_unit
    def test_remove_from_config_input_params_validation(self):
        """
        <b>Description:</b>
        Check "remove_from_config" function input parameters validation

        <b>Input data:</b>
        "remove_from_config" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "remove_from_config" function
        """
        correct_values_dict = {"config": Config(), "key": "key_1"}
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "config" parameter
            ("config", unexpected_int),
            # Unexpected integer is specified as "key" parameter
            ("key", unexpected_int),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=remove_from_config,
        )

    @e2e_pytest_unit
    def test_get_data_cfg_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_data_cfg" function input parameters validation

        <b>Input data:</b>
        "get_data_cfg" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "get_data_cfg" function
        """
        correct_values_dict = {
            "config": Config(),
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "config" parameter
            ("config", unexpected_int),
            # Unexpected integer is specified as "subset" parameter
            ("subset", unexpected_int),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=get_data_cfg,
        )

    @e2e_pytest_unit
    def test_patch_data_pipeline_input_params_validation(self):
        """
        <b>Description:</b>
        Check "patch_data_pipeline" function input parameters validation

        <b>Input data:</b>
        "patch_data_pipeline" function input parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "patch_data_pipeline" function
        """
        correct_values_dict = {
            "config": Config(),
            "data_pipeline": "otx/algorithms/classification/configs/base/data/data_pipeline.py",
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "config" parameter
            ("config", unexpected_int),
            # Unexpected integer is specified as "subset" parameter
            ("subset", unexpected_int),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=get_data_cfg,
        )

    @e2e_pytest_unit
    def test_patch_color_conversion_input_params_validation(self):
        """
        <b>Description:</b>
        Check "patch_color_conversion" function input parameters validation

        <b>Input data:</b>
        "pipeline" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "patch_color_conversion" function
        """
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "pipeline" parameter
            unexpected_int,
            # Unexpected integer is specified as nested pipeline
            [{"correct": "dictionary"}, unexpected_int],
        ]:
            with pytest.raises(ValueError):
                patch_color_conversion(pipeline=unexpected_value)  # type: ignore



"""Single Stage detector configuration of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = "./detector.py"

model = dict(
    type="SingleStageDetector",
    train_cfg=dict(
        assigner=dict(
            type="MaxIoUAssigner",
            min_pos_iou=0.0,
            ignore_iof_thr=-1,
            gt_max_assign_all=False,
        ),
        smoothl1_beta=1.0,
        allowed_border=-1,
        pos_weight=-1,
        neg_pos_ratio=3,
        debug=False,
    ),
    test_cfg=dict(
        nms=dict(type="nms", iou_threshold=0.45),
        min_bbox_size=0,
        score_thr=0.02,
        max_per_img=200,
    ),
)



"""Adapt AVARoIHead in mmaction into OTX."""
from mmaction.core.bbox import bbox2result
from mmaction.models.heads import AVARoIHead as MMAVARoIHead
from mmdet.models import HEADS as MMDET_HEADS
from torch.onnx import is_in_onnx_export


@MMDET_HEADS.register_module(force=True)
# pylint: disable=abstract-method, unused-argument,too-many-ancestors
class AVARoIHead(MMAVARoIHead):
    """AVARoIHead for OTX."""

    # TODO Remove this after changing from forward_test to onnx_export for onnx export
    def simple_test(self, x, proposal_list, img_metas, proposals=None, rescale=False, **kwargs):
        """This is temporary soluition, since otx.mmdet is differnt with latest mmdet."""
        assert self.with_bbox, "Bbox head must be implemented."

        if isinstance(x, tuple):
            x_shape = x[0].shape
        else:
            x_shape = x.shape

        assert x_shape[0] == 1, "only accept 1 sample at test mode"
        assert x_shape[0] == len(img_metas) == len(proposal_list)

        det_bboxes, det_labels = self.simple_test_bboxes(x, img_metas, proposal_list, self.test_cfg, rescale=rescale)
        if is_in_onnx_export():
            return det_bboxes, det_labels
        bbox_results = bbox2result(det_bboxes, det_labels, self.bbox_head.num_classes, thr=self.test_cfg.action_thr)
        return [bbox_results]



"""Model configuration of OCR-Lite-HRnet-18-mod2 model for Segmentation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

_base_ = [
    "../../../common/adapters/mmcv/configs/backbones/lite_hrnet_18.py",
]

model = dict(
    type="ClassIncrEncoderDecoder",
    pretrained=None,
    decode_head=dict(
        type="FCNHead",
        in_channels=[40, 80, 160, 320],
        in_index=[0, 1, 2, 3],
        input_transform="multiple_select",
        channels=40,
        kernel_size=1,
        num_convs=1,
        concat_input=False,
        dropout_ratio=-1,
        num_classes=2,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        enable_aggregator=True,
        enable_out_norm=False,
        loss_decode=[
            dict(
                type="CrossEntropyLoss",
                use_sigmoid=False,
                loss_weight=1.0,
            ),
        ],
    ),
)

load_from = "https://storage.openvinotoolkit.org/repositories/openvino_training_extensions\
/models/custom_semantic_segmentation/litehrnet18_imagenet1k_rsc.pth"

fp16 = dict(loss_scale=512.0)



"""Search space class for HPO."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import logging
import math
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

from otx.algorithms.common.utils.utils import check_positive

logger = logging.getLogger(__name__)

AVAILABLE_SEARCH_SPACE_TYPE = ["uniform", "quniform", "loguniform", "qloguniform", "choice"]


class SingleSearchSpace:
    """The class which implements single search space used for HPO.

    This class supports uniform and quantized uniform with normal and log scale
    in addition to categorical type. Quantized type has step which is unit for change.

    Args:
        type (str): type of hyper parameter in search space.
                    supported types: uniform, loguniform, quniform, qloguniform, choice
        min (float or int, optional): upper bounding of search space.
                                      If type isn't choice, this value is required.
        max (float or int, optional): lower bounding of search space
                                      If type isn't choice, this value is required.
        step (int, optional): unit for change. If type is quniform or qloguniform,
                              This value is required.
        log_base (int, optional): base of logarithm. Default value is 2.
        choice_list (list, optional): candidiates for choice type. If task is choice,
                                      this value is required.
    """

    def __init__(
        self,
        type: str,
        min: Optional[Union[float, int]] = None,
        max: Optional[Union[float, int]] = None,
        step: Optional[Union[float, int]] = None,
        log_base: Optional[int] = 2,
        choice_list: Optional[Union[List, Tuple]] = None,
    ):
        # pylint: disable=redefined-builtin
        self._type = type
        if self.is_categorical():
            self._choice_list = choice_list
            self._align_min_max_to_choice_list_if_categorical()
        else:
            self._min = min
            self._max = max
            self._step = step
            self._log_base = log_base
        self._check_all_value_is_right()

    @property
    def type(self):
        """Type of hyper parameter in search space."""
        return self._type

    @property
    def min(self):
        """Lower bounding of search space."""
        return self._min

    @property
    def max(self):
        """Upper bounding of search space."""
        return self._max

    @property
    def step(self):
        """Unit for change."""
        return self._step

    @property
    def log_base(self):
        """Base of logarithm."""
        return self._log_base

    @property
    def choice_list(self):
        """Candidiates for choice type."""
        return self._choice_list

    def set_value(
        self,
        type: Optional[str] = None,
        min: Optional[Union[float, int]] = None,
        max: Optional[Union[float, int]] = None,
        step: Optional[Union[float, int]] = None,
        log_base: Optional[int] = None,
        choice_list: Optional[Union[List, Tuple]] = None,
    ):
        # pylint: disable=redefined-builtin
        """Set attributes of the class.

        There are some dependencies when setting attributes of this class.
        So, modifying a single attribute at a time can set wrong value.
        For example, `max` can be set as a value lower than `min`.
        To prevent it, this function priovides a way to set all necessary values at a time.

        Args:
            type (str): type of hyper parameter in search space.
                        supported types: uniform, loguniform, quniform, qloguniform, choice
            min (float or int, optional): upper bounding of search space.
                                        If type isn't choice, this value is required.
            max (float or int, optional): lower bounding of search space
                                        If type isn't choice, this value is required.
            step (int, optional): unit for change. If type is quniform or qloguniform,
                                This value is required.
            log_base (int, optional): base of logarithm. Default value is 2.
            choice_list (list, optional): candidiates for choice type. If task is choice,
                                        this value is required.
        """
        if type is not None:
            self._type = type
        if min is not None:
            self._min = min
        if max is not None:
            self._max = max
        if step is not None:
            self._step = step
        if log_base is not None:
            self._log_base = log_base
        if choice_list is not None:
            self._choice_list = choice_list
        if self.is_categorical():
            self._align_min_max_to_choice_list_if_categorical()

        self._check_all_value_is_right()

    def _align_min_max_to_choice_list_if_categorical(self):
        if self.is_categorical():
            self._min = 0
            self._max = len(self._choice_list) - 1

    def _check_all_value_is_right(self):
        # pylint: disable=too-many-branches
        if self._type not in AVAILABLE_SEARCH_SPACE_TYPE:
            raise ValueError(
                f"type should be one of {', '.join(AVAILABLE_SEARCH_SPACE_TYPE)}. " f"But your argument is {self._type}"
            )

        if self.is_categorical():
            if len(self._choice_list) <= 1:
                raise ValueError("If type is choice, choice_list should have more than one element")
            if self._min != 0:
                raise ValueError("if type is categorical, min should be 0.")
            if self._max != len(self._choice_list) - 1:
                raise ValueError("if type is categorical, max should be last index number of choice_list.")
        else:
            if min is None:
                raise ValueError("If type isn't choice, you should set min value of search space.")
            if max is None:
                raise ValueError("If type isn't choice, you should set max value of search space.")

            if self._min >= self._max:
                raise ValueError(
                    "max value should be greater than min value.\n" f"max value : {self._max} / min value : {self._min}"
                )

            if self.use_log_scale():
                if self._log_base <= 1:
                    raise ValueError("log base should be greater than 1.\n" f"your log base value is {self._log_base}.")
                if self._min <= 0:
                    raise ValueError(
                        "If you use log scale, min value should be greater than 0.\n" f"your min value is {self._min}"
                    )
            if self.use_quantized_step():
                if self._step is None:
                    raise ValueError(f"The {self._type} type requires step value. But it doesn't exists")
                check_positive(self._step, "step")
                if self._step > self._max - self._min:
                    raise ValueError(
                        "Difference between min and max is greater than step.\n"
                        f"Current value is min : {self._min}, max : {self._max}, step : {self._step}"
                    )

    def __repr__(self):
        """Print serach space status."""
        if self.is_categorical():
            return f"type: {self._type}, candidiate : {', '.join(self._choice_list)}"
        rep = f"type: {self._type}, search space : {self._min} ~ {self._max}"
        if self.use_quantized_step():
            rep += f", step : {self._step}"
        if self.use_log_scale():
            rep += f", log base : {self._log_base}"
        return rep

    def is_categorical(self):
        """Check current instance is categorical type."""
        return self._type == "choice"

    def use_quantized_step(self):
        """Check current instance is one of type to use `step`."""
        return self._type in ("quniform", "qloguniform")

    def use_log_scale(self):
        """Check current instance is one of type to use `log scale`."""
        return self._type in ("loguniform", "qloguniform")

    def lower_space(self):
        """Get lower bound value considering log scale if necessary."""
        if self.use_log_scale():
            return math.log(self._min, self._log_base)
        return self._min

    def upper_space(self):
        """Get upper bound value considering log scale if necessary."""
        if self.use_log_scale():
            return math.log(self._max, self._log_base)
        return self._max

    @typing.no_type_check
    def space_to_real(self, number: Union[int, float]) -> Union[int, float]:
        """Convert search space from HPO perspective to human perspective.

        Args:
            number (Union[int, float]): Value to convert

        Returns:
            Union[int, float]: value converted to human perspective.
        """
        if self.is_categorical():
            idx = max(min(int(number), len(self._choice_list) - 1), 0)
            return self._choice_list[idx]
        if self.use_log_scale():
            number = self._log_base**number
        if self.use_quantized_step():
            gap = self._min % self._step
            number = round((number - gap) / self._step) * self._step + gap
        return number

    def real_to_space(self, number: Union[int, float]) -> Union[int, float]:
        """Convert search space from human perspective to HPO perspective.

        Args:
            number (Union[int, float]): Value to convert

        Returns:
            Union[int, float]: value converted to HPO perspective.
        """
        if self.use_log_scale():
            return math.log(number, self._log_base)  # type: ignore
        return number


class SearchSpace:
    """Class which manages all search spaces of hyper parameters to optimize.

    This class supports HPO algorithms by providing necessary functionalities.

    Args:
        search_space (dict) :
            search spaces of hyper parameters to optimize.
            arguemnt format is as bellow.
            {
                "some_hyper_parameter_name" : {
                    "param_type": type of search space of hyper parameter.
                                  supported types: uniform, loguniform, quniform,
                                  qloguniform or choice
                    # At this point, there are two available formats.
                    # first format is adding each key, value pairs depending on type as bellow.
                    "max" : upper_bound_value,
                    "min" : lower_bound_value,
                    ...
                    # available keys are max, min, step, log_base, choice_list.
                    # second format is adding a list whose key is "range" which contains
                    # necessary values mentioned above with proper order as bellow.
                    "range" (list): range of hyper parameter search space.
                                    What value at each position means is as bellow.
                                    uniform: [min, max]
                                    quniform: [min, max, step]
                                    loguniform: [min, max, log_base]
                                    qloguniform: [min, max, step, log_base]
                                    choice: [each categorical values, ...]
                    # First format is recommaneded, but you can use any format.
                    # Please refer SingleSearchSpace class comment for
                    # detailed explanation of each vlaues.
                }
                "another_hyper_parameter_name" : {...}
                ...
            }
    """

    def __init__(
        self,
        search_space: Dict[str, Dict[str, Any]],
    ):
        self.search_space: Dict[str, SingleSearchSpace] = {}

        for key, val in search_space.items():  # pylint: disable=too-many-nested-blocks
            if "range" not in val:
                val["type"] = val.pop("param_type")
                self.search_space[key] = SingleSearchSpace(**val)
            else:
                args = {"type": val["param_type"]}
                if val["param_type"] == "choice":
                    args["choice_list"] = val["range"]
                else:
                    if len(val) != 2:
                        logger.warning("If there is the range in keys, then other values are ignored.")
                    try:
                        args["min"] = val["range"][0]
                        args["max"] = val["range"][1]
                        if args["type"] == "quniform":
                            args["step"] = val["range"][2]
                        elif args["type"] == "loguniform":
                            if len(val["range"]) == 3:
                                args["log_base"] = val["range"][2]
                        elif args["type"] == "qloguniform":
                            args["step"] = val["range"][2]
                            if len(val["range"]) == 4:
                                args["log_base"] = val["range"][3]
                    except IndexError as exc:
                        raise ValueError(
                            "You should give all necessary value depending on search space type."
                            "which values are needed depending on type are as bellow."
                            "   - uniform : min value, max value"
                            "   - quniform : min value, max value, step"
                            "   - loguniform : min value, max value, log base(default 2)"
                            "   - qloguniform : min value, max value, step, log baes(default 2)"
                            "But your value is:"
                            f"  - {val['param_type']} : {', '.join([str(element) for element in val['range']])}"
                        ) from exc
                self.search_space[key] = SingleSearchSpace(**args)

    def __getitem__(self, key):
        """Get search space by key."""
        try:
            return self.search_space[key]
        except KeyError as exc:
            raise KeyError(f"There is no search space named {key}.") from exc

    def __repr__(self):
        """Print all search spaces."""
        return "\n".join(f"{key} => {val}" for key, val in self.search_space.items())

    def __iter__(self):
        """Iterate search spaces."""
        return self._search_space_generator()

    def __len__(self):
        """Number of search spaces."""
        return len(self.search_space)

    def _search_space_generator(self):
        for key in self.search_space:
            yield key

    def has_categorical_param(self):
        """Check there is a search space whose type is choice."""
        for param in self.search_space.values():
            if param.is_categorical():
                return True
        return False

    def get_real_config(self, config: Dict) -> Dict:
        """Convert search space of each config from HPO perspective to human perspective.

        Args:
            config (Dict): config to convert

        Returns:
            Dict: config converted to human perspective.
        """
        real_config = {}
        for param, value in config.items():
            real_config[param] = self[param].space_to_real(value)
        return real_config

    def get_space_config(self, config: Dict) -> Dict:
        """Convert search space of each config from human perspective to HPO perspective.

        Args:
            config (Dict): config to convert

        Returns:
            Dict: config converted to human perspective.
        """
        space_config = {}
        for param, value in config.items():
            space_config[param] = self[param].real_to_space(value)
        return space_config

    def get_bayeopt_search_space(self) -> Dict:
        """Return hyper parameter serach sapce as bayeopt library format."""
        bayesopt_space = {}
        for key, val in self.search_space.items():
            bayesopt_space[key] = (val.lower_space(), val.upper_space())

        return bayesopt_space

    def convert_from_zero_one_scale_to_real_space(self, config: Dict) -> Dict:
        """Convert search space of each config from zero one scale to human perspective.

        Args:
            config (Dict): config to convert

        Returns:
            Dict: config converted to human perspective.
        """
        for key, val in config.items():
            lower = self.search_space[key].lower_space()
            upper = self.search_space[key].upper_space()
            val = (upper - lower) * val + lower
            config[key] = val

        return self.get_real_config(config)



"""Unit Test for otx.algorithms.action.tasks.inference."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
from typing import Any, Dict

import numpy as np
import pytest
import torch
from mmcv.utils import Config
from torch import nn

from otx.algorithms.action.configs.base.configuration import ActionConfig
from otx.algorithms.action.tasks.inference import ActionInferenceTask
from otx.api.configuration import ConfigurableParameters
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.model import (
    ModelConfiguration,
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
)
from otx.api.entities.model_template import InstantiationType, TaskFamily, TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockModelTemplate,
    generate_action_cls_otx_dataset,
    generate_action_det_otx_dataset,
    generate_labels,
    return_inputs,
)


class MockModule(nn.Module):
    """Mock class for nn.Module."""

    def forward(self, inputs: Any):
        return inputs


class MockModel(nn.Module):
    """Mock class for pytorch model."""

    def __init__(self, task_type):
        super().__init__()
        self.module = MockModule()
        self.module.backbone = MockModule()
        self.backbone = MockModule()
        self.task_type = task_type

    def forward(self, return_loss: bool, imgs: DatasetItemEntity):
        forward_hooks = list(self.module.backbone._forward_hooks.values())
        for hook in forward_hooks:
            hook(1, 2, 3)
        if self.task_type == "cls":
            return np.array([[0, 0, 1]])
        return [[np.array([[0, 0, 1, 1, 0.1]]), np.array([[0, 0, 1, 1, 0.2]]), np.array([[0, 0, 1, 1, 0.7]])]]

    @staticmethod
    def named_parameters():
        return {"name": torch.Tensor([0.5])}.items()


class MockDataset(DatasetEntity):
    """Mock class for mm_dataset."""

    def __init__(self, dataset: DatasetEntity, task_type: str):
        self.dataset = dataset
        self.task_type = task_type

    def evaluate(self, prediction, *args, **kwargs):
        if self.task_type == "cls":
            return {"accuracy": 1.0}
        else:
            return {"mAP@0.5IOU": 1.0}


class MockDataLoader:
    """Mock class for data loader."""

    def __init__(self, dataset: DatasetEntity):
        self.dataset = dataset
        self.iter = iter(self.dataset)

    def __len__(self) -> int:
        return len(self.dataset)

    def __next__(self) -> Dict[str, DatasetItemEntity]:
        return {"imgs": next(self.iter)}

    def __iter__(self):
        return self


def mock_export_model(model, recipe_cfg, onnx_model_path, output_dir_path):
    """Mock function for export_model function."""

    dummy_data = np.ndarray((1, 1, 1))
    with open(os.path.join(output_dir_path, "dummpy.bin"), "wb") as f:
        f.write(dummy_data)
    with open(os.path.join(output_dir_path, "dummpy.xml"), "wb") as f:
        f.write(dummy_data)


class TestActionInferenceTask:
    """Test class for ActionInferenceTask.

    Details are explained in each test function.
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.video_len = 3
        self.frame_len = 3

        cls_labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.cls_label_schema = LabelSchemaEntity()
        cls_label_group = LabelGroup(
            name="labels",
            labels=cls_labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        self.cls_label_schema.add_group(cls_label_group)
        cls_template = MockModelTemplate(
            model_template_id="template_id",
            model_template_path="template_path",
            name="template",
            task_family=TaskFamily.VISION,
            task_type=TaskType.ACTION_CLASSIFICATION,
            instantiation=InstantiationType.CLASS,
        )
        self.cls_task_environment = TaskEnvironment(
            model=None,
            hyper_parameters=ConfigurableParameters(header="h-params"),
            label_schema=self.cls_label_schema,
            model_template=cls_template,
        )

        self.cls_dataset = generate_action_cls_otx_dataset(self.video_len, self.frame_len, cls_labels)
        self.cls_task = ActionInferenceTask(task_environment=self.cls_task_environment)

        det_labels = generate_labels(3, Domain.ACTION_DETECTION)
        self.det_label_schema = LabelSchemaEntity()
        det_label_group = LabelGroup(
            name="labels",
            labels=det_labels,
            group_type=LabelGroupType.EXCLUSIVE,
        )
        self.det_label_schema.add_group(det_label_group)
        det_template = MockModelTemplate(
            model_template_id="template_id",
            model_template_path="template_path",
            name="template",
            task_family=TaskFamily.VISION,
            task_type=TaskType.ACTION_DETECTION,
            instantiation=InstantiationType.CLASS,
        )
        self.det_task_environment = TaskEnvironment(
            model=None,
            hyper_parameters=ConfigurableParameters(header="h-params"),
            label_schema=self.det_label_schema,
            model_template=det_template,
        )

        self.det_dataset = generate_action_det_otx_dataset(self.video_len, self.frame_len, det_labels)[0]
        self.det_task = ActionInferenceTask(task_environment=self.det_task_environment)

    @e2e_pytest_unit
    # TODO Sepearate add prediction function test and infer funciton test
    def test_infer(self, mocker) -> None:
        """Test infer function.

        <Steps>
            1. Create mock model for action classification
            2. Create mock recipe for action classification
            3. Run infer funciton
            4. Check whether inference results are added to output
            5. Do 1 - 4 for action detection
        """

        mocker.patch(
            "otx.algorithms.action.tasks.inference.build_dataset", return_value=MockDataset(self.cls_dataset, "cls")
        )
        mocker.patch(
            "otx.algorithms.action.tasks.inference.build_dataloader", return_value=MockDataLoader(self.cls_dataset)
        )
        mocker.patch("otx.algorithms.action.tasks.inference.MMDataParallel", return_value=MockModel("cls"))
        self.cls_task._model = MockModel("cls")
        self.cls_task._recipe_cfg = Config(
            {
                "data": {"test": {"otx_dataset": None}, "workers_per_gpu": 1},
                "gpu_ids": [0],
                "evaluation": {"final_metric": "accuracy"},
            }
        )
        inference_parameters = InferenceParameters(is_evaluation=True)
        outputs = self.cls_task.infer(self.cls_dataset, inference_parameters)
        for output in outputs:
            assert len(output.get_annotations()[0].get_labels()) == 2

        mocker.patch(
            "otx.algorithms.action.tasks.inference.build_dataset", return_value=MockDataset(self.det_dataset, "det")
        )
        mocker.patch(
            "otx.algorithms.action.tasks.inference.build_dataloader", return_value=MockDataLoader(self.det_dataset)
        )
        mocker.patch("otx.algorithms.action.tasks.inference.MMDataParallel", return_value=MockModel("det"))
        self.det_task._model = MockModel("det")
        self.det_task._recipe_cfg = Config(
            {
                "data": {"test": {"otx_dataset": None}, "workers_per_gpu": 1},
                "gpu_ids": [0],
                "evaluation": {"final_metric": "mAP@0.5IOU"},
            }
        )
        inference_parameters = InferenceParameters(is_evaluation=True)
        outputs = self.det_task.infer(self.det_dataset, inference_parameters)
        for output in outputs:
            assert len(output.get_annotations()) == 4

    @e2e_pytest_unit
    def test_evaluate(self) -> None:
        """Test evaluate function.

        <Steps>
            1. Create model entity
            2. Create result set entity
            3. Run evaluate function with same dataset, this should give 100% accuracy
            4. Run evaluate function with empty dataset, this should give 0% accuracy
            5. Do 1 - 4 for action detection
        """
        _config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        _model = ModelEntity(self.cls_dataset, _config)
        resultset = ResultSetEntity(_model, self.cls_dataset, self.cls_dataset)
        self.cls_task.evaluate(resultset)
        assert resultset.performance.score.value == 1.0

        resultset = ResultSetEntity(_model, self.cls_dataset, self.cls_dataset.with_empty_annotations())
        self.cls_task.evaluate(resultset)
        assert resultset.performance.score.value == 0.0

        _config = ModelConfiguration(ActionConfig(), self.det_label_schema)
        _model = ModelEntity(self.det_dataset, _config)
        resultset = ResultSetEntity(_model, self.det_dataset, self.det_dataset)
        self.det_task.evaluate(resultset)
        assert resultset.performance.score.value == 0.0

    @e2e_pytest_unit
    def test_export(self, mocker) -> None:
        """Test export function.

        <Steps>
            1. Create model entity
            2. Run export function
            3. Check output model attributes
        """
        _config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        _model = ModelEntity(self.cls_dataset, _config)
        self.cls_task._model = None
        self.cls_task._recipe_cfg = None
        mocker.patch("otx.algorithms.action.tasks.inference.ActionInferenceTask._init_task", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.inference.export_model", side_effect=mock_export_model)
        self.cls_task.export(ExportType.OPENVINO, _model)

        assert _model.model_format == ModelFormat.OPENVINO
        assert _model.optimization_type == ModelOptimizationType.MO
        assert _model.get_data("openvino.bin") is not None
        assert _model.get_data("openvino.xml") is not None
        assert _model.get_data("confidence_threshold") is not None
        assert _model.precision == self.cls_task._precision
        assert _model.optimization_methods == self.cls_task._optimization_methods
        assert _model.get_data("label_schema.json") is not None

    @e2e_pytest_unit
    def test_init_task(self, mocker) -> None:
        """Test _init_task function.

        Check model is generated from _init_task function.
        """
        mocker.patch("otx.algorithms.action.tasks.inference.ActionInferenceTask._initialize", return_value=True)
        mocker.patch(
            "otx.algorithms.action.tasks.inference.ActionInferenceTask._load_model", return_value=MockModel("cls")
        )
        with pytest.raises(RuntimeError):
            self.cls_task._init_task()

        self.cls_task._recipe_cfg = Config()
        self.cls_task._init_task()
        assert isinstance(self.cls_task._model, MockModel)

    @e2e_pytest_unit
    def test_load_model(self, mocker) -> None:
        """Test _load_model function.

        Check _load_model function can run _create_model function under various situations
        """

        mocker.patch(
            "otx.algorithms.action.tasks.inference.ActionInferenceTask._create_model", return_value=MockModel("cls")
        )
        mocker.patch("otx.algorithms.action.tasks.inference.load_state_dict", return_value=True)
        mocker.patch(
            "otx.algorithms.action.tasks.inference.torch.load",
            return_value={"confidence_threshold": 0.01, "model": np.array([0.01])},
        )

        with pytest.raises(Exception):
            self.cls_task._load_model(None)

        _config = ModelConfiguration(ActionConfig(), self.cls_label_schema)
        _model_entity = ModelEntity(self.cls_dataset, _config)
        _model_entity.set_data("weights.pth", np.ndarray((1, 1, 1)).tobytes())

        self.cls_task._recipe_cfg = Config({"load_from": "weights.pth"})
        model = self.cls_task._load_model(_model_entity)
        assert isinstance(model, MockModel)

        model = self.cls_task._load_model(None)
        assert isinstance(model, MockModel)

    def test_create_model(self, mocker) -> None:
        """Test _create_model function.

        Check _create_model function can run build_model funciton under various situations
        """
        mocker.patch("otx.algorithms.action.tasks.inference.build_model", return_value=MockModel("cls"))
        mocker.patch("otx.algorithms.action.tasks.inference.load_checkpoint", return_value=True)

        _config = Config({"model": Config(), "load_from": "weights.pth"})
        model = self.cls_task._create_model(_config, False)
        assert isinstance(model, MockModel)
        model = self.cls_task._create_model(_config, True)
        assert isinstance(model, MockModel)

    def test_unload(self) -> None:
        """Test unload function."""
        self.cls_task.unload()

    def test_init_recipe_hparam(self, mocker) -> None:
        """Test _init_recipe_hparam function."""

        mocker.patch(
            "otx.algorithms.action.tasks.inference.BaseTask._init_recipe_hparam",
            return_value=Config(
                {"data": {"samples_per_gpu": 4}, "lr_config": {"warmup_iters": 3}, "runner": {"max_epochs": 10}}
            ),
        )

        self.cls_task._recipe_cfg = Config({"lr_config": Config()})
        out = self.cls_task._init_recipe_hparam()

        assert self.cls_task._recipe_cfg.lr_config.warmup == "linear"
        assert self.cls_task._recipe_cfg.lr_config.warmup_by_epoch is True
        assert self.cls_task._recipe_cfg.total_epochs == 10
        assert out.data.videos_per_gpu == 4
        assert out.use_adaptive_interval == self.cls_task._hyperparams.learning_parameters.use_adaptive_interval

    def test_init_recipe(self, mocker) -> None:
        """Test _init_recipe funciton."""

        mocker.patch("otx.algorithms.action.tasks.inference.Config.fromfile", side_effect=return_inputs)
        mocker.patch("otx.algorithms.action.tasks.inference.patch_config", return_value=True)
        mocker.patch("otx.algorithms.action.tasks.inference.set_data_classes", return_value=True)

        self.cls_task._init_recipe()
        recipe_root = os.path.abspath(os.path.dirname(self.cls_task.template_file_path))
        assert self.cls_task._recipe_cfg == os.path.join(recipe_root, "model.py")

    def test_init_model_cfg(self, mocker) -> None:
        """Test _init_model_cfg function."""

        mocker.patch("otx.algorithms.action.tasks.inference.Config.fromfile", side_effect=return_inputs)

        model_cfg = self.cls_task._init_model_cfg()
        assert model_cfg == os.path.join(self.cls_task._model_dir, "model.py")



import os

import pytest

from otx.algorithms.segmentation.adapters.mmseg.utils.builder import build_segmentor
from otx.mpa.deploy.apis import NaiveExporter
from otx.mpa.exporter_mixin import ExporterMixin
from otx.mpa.seg.exporter import SegExporter
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_RECIPE_CONFIG_PATH,
    DEFAULT_SEG_TEMPLATE_DIR,
)


class TestOTXSegExporter:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(DEFAULT_RECIPE_CONFIG_PATH)
        self.exporter = SegExporter(name="", mode="train", config=cfg, common_cfg=None, index=0)
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_run(self, mocker):
        args = {"precision": "FP32", "model_builder": build_segmentor}
        mocker.patch.object(ExporterMixin, "run", return_value=True)
        returned_value = self.exporter.run(self.model_cfg, "", self.data_cfg, **args)

        assert "model_builder" in args
        assert returned_value is True

    @e2e_pytest_unit
    def test_naive_export(self, mocker):
        mock_export_ov = mocker.patch.object(NaiveExporter, "export2openvino")
        self.exporter.naive_export("", build_segmentor, "FP32", self.data_cfg)

        mock_export_ov.assert_called_once()



"""Unit Test for otx.algorithms.action.adapters.mmaction.data.cls_dataset."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
import torch
from mmcv.utils import ConfigDict
from mmdet.models.detectors.faster_rcnn import FasterRCNN
from torch import nn

from otx.algorithms.action.adapters.mmaction.models.detectors.fast_rcnn import (
    AVAFastRCNN,
    ONNXPool3D,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockDetector(nn.Module):
    """Mock class for person detector."""

    def onnx_export(self, *args, **kwargs):
        """Return dummy person detection results."""

        sample_det_bboxes = torch.Tensor([[0.0, 0.0, 1.0, 1.0, 1.0]] * 100).unsqueeze(0)
        sample_det_labels = torch.ones(1, 100)
        sample_det_labels[0][0] = 0
        return sample_det_bboxes, sample_det_labels


class TestONNXPool3d:
    """Test ONNXPool3D class.

    1. Check every possible ONNXPool3D generation
    2. Check every possible ONNXPool3D actuall pooling input tensor as expected
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.temporal_avg_pool = ONNXPool3D("temporal", "avg")
        self.temporal_max_pool = ONNXPool3D("temporal", "max")
        self.spatial_avg_pool = ONNXPool3D("spatial", "avg")
        self.spatial_max_pool = ONNXPool3D("spatial", "max")

    @e2e_pytest_unit
    def test_init(self) -> None:
        """Test __init__ function."""

        self.temporal_avg_pool.pool == nn.functional.avg_pool3d
        self.temporal_max_pool.pool == nn.functional.max_pool3d
        self.spatial_avg_pool.pool == nn.functional.avg_pool3d
        self.spatial_max_pool.pool == nn.functional.max_pool3d

    @e2e_pytest_unit
    def test_forward(self) -> None:
        """Test forward function."""

        sample_input = torch.randn(1, 100, 8, 8, 8)
        output = self.temporal_avg_pool(sample_input)
        assert list(output.shape) == [1, 100, 1, 8, 8]
        output = self.temporal_max_pool(sample_input)
        assert list(output.shape) == [1, 100, 1, 8, 8]
        output = self.spatial_avg_pool(sample_input)
        assert list(output.shape) == [1, 100, 8, 1, 1]
        output = self.spatial_max_pool(sample_input)
        assert list(output.shape) == [1, 100, 8, 1, 1]


class TestAVAFastRCNN:
    """Test AVAFastRCNN class.

    1. Check add_detector function
    2. Check patch_pools function
    3. Check forward_infer function
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.model = AVAFastRCNN(
            backbone=ConfigDict(type="X3D", gamma_w=1, gamma_b=2.25, gamma_d=2.2),
            roi_head=ConfigDict(
                type="AVARoIHead",
                bbox_roi_extractor=ConfigDict(
                    type="SingleRoIExtractor3D", roi_layer_type="RoIAlign", output_size=8, with_temporal_pool=True
                ),
                bbox_head=ConfigDict(
                    type="BBoxHeadAVA", in_channels=432, num_classes=81, multilabel=False, dropout_ratio=0.5
                ),
            ),
            train_cfg=ConfigDict(
                rcnn=ConfigDict(
                    assigner=ConfigDict(type="MaxIoUAssignerAVA", pos_iou_thr=0.9, neg_iou_thr=0.9, min_pos_iou=0.9),
                    sampler=ConfigDict(
                        type="RandomSampler", num=32, pos_fraction=1, neg_pos_ub=-1, add_gt_as_proposals=True
                    ),
                    pos_weight=1.0,
                    debug=False,
                )
            ),
            test_cfg=ConfigDict(rcnn=ConfigDict(action_thr=0.002)),
        )

    @e2e_pytest_unit
    def test_add_detector(self, mocker) -> None:
        """Test add_deector function.

        <Steps>
            1. Check added detector is FasterRCNN
            2. Check added detector has COCO classes
            3. Check added detector's CLASSES is properly initialized
            4. Check added detector raise exception if detector's first class is not person
        """

        mocker.patch(
            "otx.algorithms.action.adapters.mmaction.models.detectors.fast_rcnn.load_checkpoint",
            return_value={"meta": {"CLASSES": ["person", "motorcycle", "car"]}},
        )
        self.model.add_detector()
        assert isinstance(self.model.detector, FasterRCNN)
        assert self.model.detector.roi_head.bbox_head.num_classes == 80
        assert self.model.detector.CLASSES == ["person", "motorcycle", "car"]

        mocker.patch(
            "otx.algorithms.action.adapters.mmaction.models.detectors.fast_rcnn.load_checkpoint",
            return_value={"meta": {"CLASSES": ["motorcycle", "car", "person"]}},
        )
        with pytest.raises(Exception):
            self.model.add_detector()

    @e2e_pytest_unit
    def test_patch_pools(self) -> None:
        """Test patch_pools function.

        <Steps>
            1. Check bbox_head's temporal pool is avg_pool and it pools through temporal axis
            2. Check bbox_head's spatial pool is max_pool and it pools through spatial axis
        """

        self.model.patch_pools()
        assert isinstance(self.model.roi_head.bbox_head.temporal_pool, ONNXPool3D)
        assert self.model.roi_head.bbox_head.temporal_pool.pool == nn.functional.avg_pool3d
        assert self.model.roi_head.bbox_head.temporal_pool.dim == "temporal"
        assert isinstance(self.model.roi_head.bbox_head.spatial_pool, ONNXPool3D)
        assert self.model.roi_head.bbox_head.spatial_pool.pool == nn.functional.max_pool3d
        assert self.model.roi_head.bbox_head.spatial_pool.dim == "spatial"

    @e2e_pytest_unit
    def test_forward_infer(self, mocker) -> None:
        """Test forward_infer function.

        <Steps>
            1. Prepare sample imgs and img_metas
            2. Patch model's detector to MockDetector
            3. Check amount of output bboxes and output labels are same
            4. Check output bboxes have 4 cooridnates
            5. Check output labels have all num_classes of bbox_head
        """

        mocker.patch(
            "otx.algorithms.action.adapters.mmaction.models.heads.roi_head.is_in_onnx_export", return_value=True
        )
        sample_imgs = [torch.randn(1, 3, 32, 256, 256)]
        sample_img_metas = [
            [
                {
                    "img_shape": torch.Tensor((256, 256)),
                    "ori_shape": (192, 256),
                    "pad_shape": (256, 256),
                    "filename": "demo_vid.png",
                    "scale_factor": np.array([256 / 256, 256 / 192, 256 / 256, 256 / 192]),
                    "flip": False,
                    "show_img": False,
                    "flip_direction": None,
                }
            ]
        ]
        self.model.detector = MockDetector()
        with torch.no_grad():
            bboxes, labels = self.model.forward_infer(sample_imgs, sample_img_metas)
        assert bboxes.shape[-1] == 4
        assert labels.shape[-1] == self.model.roi_head.bbox_head.num_classes
        assert bboxes.shape[0] == labels.shape[0]



"""NNCF for mmcv."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa

from .utils import get_fake_input, model_eval, wrap_nncf_model

__all__ = [
    "get_fake_input",
    "model_eval",
    "wrap_nncf_model",
]



_base_ = ["./pipelines/twocrop_pipeline.py"]

__dataset_type = "OTXClsDataset"

__train_pipeline = {{_base_.train_pipeline}}
__test_pipeline = {{_base_.test_pipeline}}

__samples_per_gpu = 16

data = dict(
    samples_per_gpu=__samples_per_gpu,
    workers_per_gpu=2,
    train=dict(type=__dataset_type, pipeline=__train_pipeline),
    val=dict(type=__dataset_type, test_mode=True, pipeline=__test_pipeline),
    test=dict(type=__dataset_type, test_mode=True, pipeline=__test_pipeline),
)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import tempfile

from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import (
    create_dataloader,
    create_nncf_runner,
)


class TestAccuracyAwareRunner:
    @e2e_pytest_unit
    def test_run(self):
        dataloader = create_dataloader()

        with tempfile.TemporaryDirectory() as tempdir:
            runner = create_nncf_runner(tempdir)
            runner.run([dataloader])
            assert [f for f in os.listdir(tempdir) if f.startswith("best_accuracy")]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import inspect
from typing import Dict, List, Optional, Union

import torch

from ..op import Operation


class OperationModule(torch.nn.Module):
    def __init__(
        self,
        op: Operation,
        dependent_ops: Union[List[Optional[Operation]], Dict[str, Optional[Operation]]],
    ):
        super().__init__()

        self.op = op
        self._dependent_ops = torch.nn.ModuleDict()

        spec = inspect.getfullargspec(op.forward)
        kwargs = spec.args[1:]

        self._dependents_with_defaults = []
        if spec.defaults:
            self._dependents_with_defaults = spec.args[-len(spec.defaults) :]

        if isinstance(dependent_ops, list):
            assert len(dependent_ops) == len(kwargs)
            for op, kwarg in zip(dependent_ops, kwargs):
                self._dependent_ops[kwarg] = op
        elif isinstance(dependent_ops, dict):
            for kwarg in kwargs:
                self._dependent_ops[kwarg] = dependent_ops[kwarg]
        else:
            raise NotImplementedError

    def forward(self, *args, **kwargs):
        inputs = {k: v() if v is not None else None for k, v in self._dependent_ops.items()}

        if args:
            empty_input_keys = [k for k, v in self._dependent_ops.items() if v is None]
            for key, val in zip(empty_input_keys, args):
                inputs[key] = val
        if kwargs:
            for key, val in kwargs.items():
                if inputs[key] is not None:
                    raise ValueError(f"duplicated key {key}")
                inputs[key] = val

        assert all(v is not None for v in inputs.values() if v not in self._dependents_with_defaults)

        return self.op(**inputs)

    @property
    def type(self):
        return self.op.type

    @property
    def version(self):
        return self.op.version

    @property
    def name(self):
        return self.op.name

    @property
    def shape(self):
        return self.op.shape

    @property
    def attrs(self):
        return self.op.attrs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.runner import HOOKS, Fp16OptimizerHook


@HOOKS.register_module()
class Fp16SAMOptimizerHook(Fp16OptimizerHook):
    """Sharpness-aware Minimization optimizer hook

    Implemented as OptimizerHook for MMCV Runners
    - Paper ref: https://arxiv.org/abs/2010.01412
    - code ref: https://github.com/davda54/sam
    """

    def __init__(self, rho=0.05, start_epoch=1, **kwargs):
        super().__init__(**kwargs)
        self.rho = rho
        self.start_epoch = start_epoch
        if rho < 0.0:
            raise ValueError("rho should be greater than 0 for SAM optimizer")

    def after_train_iter(self, runner):
        """Perform SAM optimization
        0. compute current loss (DONE IN model.train_step())
        1. compute current gradient
        2. move param to the approximate local maximum: w + e(w) = w + rho*norm_grad
        3. compute maximum loss
        4. compute SAM gradient on maximum loss
        5. restore parram to original param
        6. update param using SAM gradient

        Assuming model.current_batch had been set in model.train_step()
        """
        current_batch = self._get_current_batch(runner.model)
        if current_batch is None or runner.epoch + 1 < self.start_epoch:
            # Apply original parameter update
            return super().after_train_iter(runner)

        # Current gradient
        runner.optimizer.zero_grad()
        curr_loss = runner.outputs["loss"]
        curr_loss.backward()

        # Move to local maximum
        param2move = {}
        with torch.no_grad():
            scale = self.rho / (self._grad_norm(runner.optimizer) + 1e-12)
            for param_group in runner.optimizer.param_groups:
                for param in param_group["params"]:
                    if param.grad is None:
                        continue
                    e_w = param.grad * scale.to(param)
                    param.add_(e_w)  # Climb toward gradient (increasing loss)
                    param2move[param] = e_w  # Saving for param restore

        # SAM gradient
        runner.optimizer.zero_grad()
        max_outputs = runner.model.train_step(current_batch, runner.optimizer)  # forward() on maxima param
        max_loss = max_outputs["loss"]
        self.loss_scaler.scale(max_loss).backward()
        self.loss_scaler.unscale_(runner.optimizer)

        # Restore to original param
        with torch.no_grad():
            for param_group in runner.optimizer.param_groups:
                for param in param_group["params"]:
                    if param.grad is None:
                        continue
                    param.sub_(param2move[param])  # Down to original param

        # Shaprpness-aware param update
        self.loss_scaler.step(runner.optimizer)  # param -= lr * sam_grad
        self.loss_scaler.update(self._scale_update_param)

        runner.meta.setdefault("fp16", {})["loss_scaler"] = self.loss_scaler.state_dict()
        runner.log_buffer.update({"sharpness": float(max_loss - curr_loss), "max_loss": float(max_loss)})

    def _get_current_batch(self, model):
        if hasattr(model, "module"):
            model = model.module
        return getattr(model, "current_batch", None)

    def _grad_norm(self, optimizer):
        # put everything on the same device, in case of model parallelism
        shared_device = optimizer.param_groups[0]["params"][0].device
        norm = torch.norm(
            torch.stack(
                [
                    p.grad.norm(p=2).to(shared_device)
                    for group in optimizer.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p=2,
        )
        return norm



"""Test L2SP loss with ignore."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from tempfile import TemporaryDirectory

import numpy as np
import pytest
import torch
from torch import nn
from torchvision.models import resnet18

from otx.mpa.modules.models.losses.l2sp_loss import L2SPLoss


class DummyModel(nn.Module):
    """Creates a dummy model for testing purposes."""

    def __init__(self, set_init_weights: float = 0.0):
        super().__init__()
        self.fc = nn.Linear(4, 1)
        self.cnn = nn.Conv2d(1, 1, 1)
        self._set_weights(set_init_weights)

    def _set_weights(self, set_init_weights: float = 0.0):
        """Sets weights of all the parameters with a given value."""
        for param in self.parameters():
            param.data.fill_(set_init_weights)


class TestL2SPLoss:
    """Test L2SP loss."""

    @pytest.fixture(scope="class")
    def model_checkpoint(self):
        """Create a model checkpoint."""
        with TemporaryDirectory() as tmp_path:
            model = DummyModel(set_init_weights=1.0)
            model_ckpt = tmp_path + "/model.ckpt"
            torch.save(model.state_dict(), model_ckpt)
            yield model_ckpt

    def test_loss(self, model_checkpoint):
        """Test loss output."""
        new_model = DummyModel(set_init_weights=0.0)
        loss = L2SPLoss(new_model, model_checkpoint)

        assert np.round(loss().item(), 4).item() == 0.0007

        new_model = DummyModel(set_init_weights=1.0)
        loss = L2SPLoss(new_model, model_checkpoint)
        assert loss().item() == 0.0

    def test_value_error(self, model_checkpoint):
        """Test ValueError."""
        model = DummyModel()

        with pytest.raises(ValueError):
            L2SPLoss(model, model_ckpt=None)

        with pytest.raises(ValueError):
            L2SPLoss(model, model_checkpoint, loss_weight=-1.0)

    def test_state_dict(self):
        """Test model load with weights stored under state_dict key."""
        with TemporaryDirectory() as tmp_path:
            model = DummyModel(set_init_weights=1.0)
            model_ckpt = tmp_path + "/model.ckpt"
            torch.save({"state_dict": model.state_dict()}, model_ckpt)
            new_model = DummyModel(set_init_weights=1.0)
            loss = L2SPLoss(new_model, model_ckpt)
            assert loss().item() == 0.0

    def test_backbone(self):
        """Test model load with backbone weights.

        This tests the case when the new model has backbone param whereas the source model as the backbone layers
        directly.
        """
        with TemporaryDirectory() as tmp_path:
            model = DummyModel()
            for param, value in resnet18(pretrained=True).named_children():
                setattr(model, param, value)
            model_ckpt = tmp_path + "/model.ckpt"
            torch.save(model.state_dict(), model_ckpt)

            new_model = DummyModel()
            setattr(new_model, "backbone", resnet18(pretrained=True))
            # new_model.load_state_dict(torch.load(model_ckpt))
            loss = L2SPLoss(new_model, model_ckpt)
            assert loss().item() == 0.0



"""This module implements the AnalyseParameters entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from dataclasses import dataclass
from typing import Any, Callable, Optional


# pylint: disable=unused-argument
def default_progress_callback(progress: int, score: Optional[float] = None):
    """This is the default progress callback for OptimizationParameters."""


@dataclass
class InferenceParameters:
    """Inference parameters.

    Attributes:
        is_evaluation: Set to ``True`` if the output dataset is intended
            to be used for evaluation purposes. In this scenario, any
            postprocessing filtering (such as thresholding and NMS)
            should be disabled to avoid interfering with algorithms such
            as NMS.
        update_progress: Callback which can be used to provide updates
            about the progress of a task.
        explainer: Explain algorithm to be used in explanation mode.
            Will be converted automatically to lowercase.
        process_saliency_maps: Process saliency map to input image resolution and apply colormap
        explain_predicted_classes: If set to True, provide explanations only for predicted classes.
            Otherwise, explain all classes.
    """

    is_evaluation: bool = False
    update_progress: Callable[[int, Optional[float]], Any] = default_progress_callback

    # TODO(negvet): use separate ExplainParameters dataclass for this
    explainer: str = ""
    process_saliency_maps: bool = False
    explain_predicted_classes: bool = True



"""Unit Test for otx.algorithms.detection.adapters.openvino.model_wrappers.openvino_models."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict

import numpy as np
import pytest
from mmcv.utils import Config
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter

from otx.algorithms.detection.adapters.openvino.model_wrappers.openvino_models import (
    BatchBoxesLabelsParser,
    OTXMaskRCNNModel,
    OTXSSDModel,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockOpenvinoAdapter(OpenvinoAdapter):
    """Mock class for OpenvinoAdapter."""

    def __init__(self):
        pass


class MockBatchBoxesLabelsParser(BatchBoxesLabelsParser):
    """Mock class for BatchBoxesLabelsParser."""

    def __init__(self):
        self.labels_layer = "labels"
        self.bboxes_layer = "boxes"
        self.input_size = (10, 10)


class MockOTXMaskRCNNModel(OTXMaskRCNNModel):
    """Mock class for OTXMaskRCNNModel."""

    def __init__(self, *args):
        self.inputs: Dict[str, np.ndarray] = {
            "image": np.ndarray([1, 3, 10, 10]),
        }

        self.outputs: Dict[str, Config] = {
            "boxes": Config({"names": "boxes", "shape": [1, 1, 5]}),
            "labels": Config({"names": "labels", "shape": [1, 1]}),
            "masks": Config({"names": "masks", "shape": [1, 0, 28, 28]}),
            "feature_vector": Config({"names": "feature_vector", "shape": [1, 1, 1, 1]}),
            "saliency_map": Config({"names": "saliency_map", "shape": [1, 1, 1]}),
        }
        self.is_segmentoly = len(self.inputs) == 2
        self.output_blob_name = self._get_outputs()
        self.confidence_threshold = 0.5
        super().__init__(MockOpenvinoAdapter)


class MockOTXSSDModel(OTXSSDModel):
    """Mock class for OTXSSDModel."""

    def __init__(self, *args):
        self.inputs: Dict[str, np.ndarray] = {
            "image": np.ndarray([1, 3, 10, 10]),
        }

        self.outputs: Dict[str, Config] = {
            "boxes": Config({"names": "boxes", "shape": [1, 1, 5]}),
            "labels": Config({"names": "labels", "shape": [1, 1]}),
            "masks": Config({"names": "masks", "shape": [1, 0, 28, 28]}),
            "feature_vector": Config({"names": "feature_vector", "shape": [1, 1, 1, 1]}),
            "saliency_map": Config({"names": "saliency_map", "shape": [1, 1, 1]}),
        }
        self.confidence_threshold = 0.375
        self.resize_type = "standard"
        self.output_parser = MockBatchBoxesLabelsParser()
        super().__init__(MockOpenvinoAdapter)


class TestOTXMaskRCNNModel:
    """Test OTXMaskRCNNModel class.

    Test postprocess function
    <Steps>
        1. Generate sample output & meta
        2. Check whether postprocess function returns (scores, classes, boxes, resized_masks) tuple with length 4.
    """

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "openvino.model_zoo.model_api.models.MaskRCNNModel.__init__",
            return_value=True,
        )
        self.model = MockOTXMaskRCNNModel()

    @e2e_pytest_unit
    def test_postprocess(self) -> None:
        """Test postprocess function."""

        sample_output = {
            "boxes": np.random.rand(1, 1, 5),
            "labels": np.random.rand(1, 1),
            "masks": np.random.rand(1, 1, 28, 28),
            "feature_vector": np.random.rand(1, 1, 1, 1),
            "saliency_map": np.random.rand(1, 1, 21),
        }
        sample_meta = {"original_shape": (10, 10, 3), "resized_shape": (5, 5, 3)}
        out = self.model.postprocess(sample_output, meta=sample_meta)
        assert len(out) == 4


class TestOTXSSDModel:
    """Test OTXSSDModel class.
    Test postprocess function
    <Steps>
        1. Generate sample output & meta
        2. Check whether postprocess function returns 'detection' with length 1.
    """

    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "otx.algorithms.detection.adapters.openvino.model_wrappers.openvino_models.OTXSSDModel.__init__",
            return_value=True,
        )
        self.model = MockOTXSSDModel()

    @e2e_pytest_unit
    def test_postprocess(self) -> None:
        """Test postprocess function."""

        sample_output = {
            "boxes": np.random.rand(1, 1, 5),
            "labels": np.random.rand(1, 1),
            "feature_vector": np.random.rand(1, 1, 1, 1),
            "saliency_map": np.random.rand(1, 1, 21),
        }
        sample_meta = {"original_shape": (10, 10, 3), "resized_shape": (5, 5, 3)}
        out = self.model.postprocess(sample_output, meta=sample_meta)
        assert len(out) <= 1


class TestBatchBoxesLabelsParser:
    """Test BatchBoxesLabelsParser class.

    1. Test __init__ function
    <Steps>
        1. Check parser's attributes for bboxes_layer and input_size
    3. Test layer_bboxes_output function
    <Steps>
        2. Check postprocess function's output return's argmax
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.layers: Dict[str, Config] = {
            "boxes": Config({"names": "boxes", "shape": [1, 1, 5]}),
            "labels": Config({"names": "labels", "shape": [1, 1]}),
            "masks": Config({"names": "masks", "shape": [1, 0, 28, 28]}),
            "feature_vector": Config({"names": "feature_vector", "shape": [1, 1, 1, 1]}),
            "saliency_map": Config({"names": "saliency_map", "shape": [1, 1, 1]}),
        }
        input_size = (10, 10)
        self.parser = BatchBoxesLabelsParser(self.layers, input_size)

    @e2e_pytest_unit
    def test_init(self) -> None:
        assert hasattr(self.parser, "bboxes_layer")
        assert hasattr(self.parser, "input_size")

    @e2e_pytest_unit
    def test_layer_bboxes_output(self) -> None:
        """Test postprocess function."""
        out = self.parser.find_layer_bboxes_output(self.layers)
        assert out == "boxes"
